"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality for visualising images.

The module provides functionality for adjusting the intensity of an image. It
provides a wrapper of the `matplotlib.pyplot.imshow` function that may exploit
the provided functions for adjusting the image intensity. Also it include a
function may be used to display a set of related images using a common
colormapping.

Routine listings
----------------
imshow(X, ax=None, intensity_func=None, intensity_args=(), \*\*kwargs)
    Function that may be used to display an image.
imsubplot(imgs, rows, titles=None, x_labels=None, y_labels=None,
    x_ticklabels=None, y_ticklabels=None, cbar_label=None,
    normalise=True, fixed_clim=None, \*\*kwargs)
    Function that may be used to display a set of related images.
mask_img_from_coords(img, coords)
    Function for masking certain parts of an image based on coordinates.
shift_mean(x_mod, x_org)
    Function for shifting mean intensity of an image based on another image.
stretch_image(img, max_val, min_val=0)
    Function for stretching the intensity of an image.

"""

from __future__ import division

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from magni.utils import plotting as _plotting
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric


def imshow(X, ax=None, intensity_func=None, intensity_args=(),
           show_axis='frame', **kwargs):
    """
    Display an image.

    Wrap `matplotlib.pyplot.imshow` to display a possibly intensity manipulated
    version of the image `X`.

    Parameters
    ----------
    X : ndarray
        The image to be displayed.
    ax : matplotlib.axes.Axes, optional
        The axes on which the image is displayed (the default is None, which
        implies that the current axes is used).
    intensity_func : FunctionType, optional
        The handle to the function used to manipulate the image intensity
        before the image is displayed (the default is None, which implies that
        no intensity manipulation is used).
    intensity_args : list or tuple, optional
        The arguments that are passed to the `intensity_func` (the default is
        (), which implies that no arguments are passed).
    show_axis : {'none', 'top', 'inherit', 'frame'}
        How the x- and y-axis are display. If 'none', no axis are displayed. If
        'top', the x-axis is displayed at the top of the image. If 'inherit',
        the axis display is inherited from `matplotlib.pyplot.imshow`. If
        'frame' only the frame is shown and not the ticks.

    Returns
    -------
    im_out : matplotlib.image.AxesImage
        The AxesImage returned by matplotlibs imshow.

    See Also
    --------
    matplotlib.pyplot.imshow : Matplotlib's imshow function.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.visualisation import imshow
    >>> X = np.arange(4).reshape(2, 2)
    >>> add_k = lambda X, k: X + k
    >>> im_out = imshow(X, intensity_func=add_k, intensity_args=(2,))

    """

    @_decorate_validation
    def validate_input():
        _numeric('X', ('boolean', 'integer', 'floating'), shape=(-1, -1))
        _generic('ax', mpl.axes.Axes, ignore_none=True)
        _generic('intensity_func', 'function', ignore_none=True)
        _generic('intensity_args', 'explicit collection')
        _generic('show_axis', 'string', value_in=('none', 'top', 'inherit',
                                                  'frame'))

    validate_input()

    _plotting.setup_matplotlib()

    # Handle ax keyword argument
    ca = plt.gca()
    if ax is not None:
        plt.sca(ax)

    # Intensity manipulation
    if intensity_func is not None:
        im_out = plt.imshow(intensity_func(X, *intensity_args), **kwargs)
    else:
        im_out = plt.imshow(X, **kwargs)

    # Display of axis
    axes = plt.gca()
    if show_axis == 'none':
        axes.axis('off')
    elif show_axis == 'top':
        axes.xaxis.tick_top()
        axes.xaxis.set_label_position('top')
    elif show_axis == 'frame':
        xlabels = axes.get_xticklabels()
        ylabels = axes.get_yticklabels()
        empty_xlabels = ['']*len(xlabels)
        empty_ylabels = ['']*len(ylabels)
        axes.set_xticklabels(empty_xlabels)
        axes.set_yticklabels(empty_ylabels)
        axes.tick_params(length=0)

    plt.sca(ca)

    return im_out


def imsubplot(imgs, rows, titles=None, x_labels=None, y_labels=None,
              x_ticklabels=None, y_ticklabels=None, cbar_label=None,
              normalise=True, fixed_clim=None, **kwargs):
    """
    Display a set of related images as subplots in a figure.

    The images `imgs` are shown in a figure with a subplot layout based on the
    number of `rows`. The `titles`, `x_labels`, `y_labels`, `x_ticklabels`,
    and `y_ticklabels` are shown in the subplots. If `normalise` is True, all
    the images will share the same normalised colorbar/colormapping i.e. a
    particular colour will correspond to the same value across all images.

    Parameters
    ----------
    imgs : list or tuple
        The images (as ndarrays) that is to be displayed.
    rows : int
        The number of rows to use in the subplot layout.
    titles : list or tuple
        The titles (as strings) to use for each of the subplots (the default is
        None, which implies that no titles are displayed).
    x_labels : list or tuple
        The x_labels (as strings) to use for each of the subplots (the default
        is None, which implies that no x_labels are displayed).
    y_labels : list or tuple
        The y_labels (as strings) to use for each of the subplots (the default
        is None, which implies that no y_labels are displayed).
    x_ticklabels : list or tuple
        The x_ticklabels (as strings or lists of strings) to use for the
        subplots (the default is None, which implies that no x_ticklabels are
        displayed).
    y_ticklabels : list or tuple
        The y_ticklabels (as strings or lists of strings) to use for the
        subplots (the default is None, which implies that no y_ticklabels are
        displayed).
    cbar_label : str
        The colorbar label to use with a normalised colormapping (the default
        is None, which implies that no colorbar label is displayed).
    fixed_clim : list or tuple
        The colorbar limits as a (min, max) sequence (the default is None,
        which implies that the colorbar limits are inferred from the data).
    normalise : bool
        The flag that indicates whether to use a normalised colormapping.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The resulting figure instance.

    See Also
    --------
    matplotlib.pyplot.subplots : Underlying subplot function.

    Notes
    -----
    The `x_ticklabels` and `y_ticklabels` may be either a collection of strings
    or a collections of collections of strings depending on wherther the labels
    should be shared across all subplots or different labels are to be used for
    each subplot.

    Additional kwargs given to the function will be passed to the underlying
    suplot instantiation function `matplotlib.pyplot.subplots`.

    If `normalise` is True, the common colorbar is shown below the subplots.

    The implementation of the normalisation feature is based on the Pylab
    example: http://matplotlib.org/examples/pylab_examples/multi_image.html.

    Examples
    --------
    For example, show two images next to each other with a common colormapping:

    >>> import numpy as np
    >>> from magni.imaging.visualisation import imsubplot
    >>> img1 = np.arange(4).reshape(2, 2)
    >>> img2 = np.ones((4, 4))
    >>> fig = imsubplot([img1, img2], 1, titles=['arange', 'ones'],
    ... x_labels=['x1', 'x2'], y_labels=['y1', 'y2'], cbar_label='Example',
    ... normalise=True)

    or show the same images with shared ticklabels and fixed colorbar limits:

    >>> common_x_ticklabels = ['a', 'b']
    >>> common_y_ticklabels = ['c', 'd']
    >>> fig = imsubplot([img1, img2], 1, x_ticklabels=common_x_ticklabels,
    ... y_ticklabels=common_y_ticklabels, fixed_clim=(0, 2), normalise=True)

    or show the same images with different colormappings and ticklabels:

    >>> x_ticklabels = [['a', 'b'], ['aa', 'bb']]
    >>> y_ticklabels = [['c', 'd'], ['cc', 'dd']]
    >>> fig = imsubplot([img1, img2], 1, x_ticklabels=x_ticklabels,
    ... y_ticklabels=y_ticklabels, normalise=False)

    """

    @_decorate_validation
    def validate_input():
        _levels('imgs', (_generic(None, 'explicit collection'),
                         _numeric(None, ('boolean', 'integer', 'floating'),
                                  shape=(-1, -1))))
        _numeric('rows', 'integer', range_='[1;inf)')
        _levels('titles', (_generic(None, 'explicit collection',
                                    len_=len(imgs), ignore_none=True),
                           _generic(None, 'string')))
        _levels('x_labels', (_generic(None, 'explicit collection',
                                      len_=len(imgs), ignore_none=True),
                             _generic(None, 'string')))
        _levels('y_labels', (_generic(None, 'explicit collection',
                                      len_=len(imgs), ignore_none=True),
                             _generic(None, 'string')))
        _levels('x_ticklabels', (_generic(None, 'explicit collection',
                                          ignore_none=True),
                                 _generic(None,
                                          ('explicit collection', 'string')),
                                 _generic(None, 'string')))
        _levels('y_ticklabels', (_generic(None, 'explicit collection',
                                          ignore_none=True),
                                 _generic(None,
                                          ('explicit collection', 'string')),
                                 _generic(None, 'string')))
        _generic('cbar_label', 'string', ignore_none=True)
        _levels('fixed_clim', (_generic(None, 'explicit collection', len_=2,
                                        ignore_none=True),
                               _numeric(None, ('integer', 'floating'))))
        _numeric('normalise', 'boolean')

    validate_input()

    cols = max(1, np.int(np.ceil(len(imgs) / rows)))
    fig, axes = plt.subplots(rows, cols, squeeze=False, **kwargs)
    axs = axes.ravel()

    vmin = 1e40
    vmax = -1e40
    ims = []
    for k, img in enumerate(imgs):
            # Show image
            ims.append(imshow(img, ax=axs[k]))

            # Handle titles, labels, etc.
            fig_strings = (titles, x_labels, y_labels)
            handles = (axs[k].set_title, axs[k].set_xlabel, axs[k].set_ylabel)
            for fig_string, handle in zip(fig_strings, handles):
                if fig_string is not None:
                    handle(fig_string[k])

            _handle_ticklabels(axs[k], k, x_ticklabels, y_ticklabels)

            # Track minimum and maximum data values
            vmin = min(vmin, np.amin(img))  # Find common minimum
            vmax = max(vmax, np.amax(img))  # Find common maximum

    if normalise:
        # Connect a Tracker to each image in order to update colormap limits
        # This way all images share common colormap limits
        common_norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
        ims[0].set_norm(common_norm)
        for k in range(1, len(ims)):
            ims[k].set_norm(common_norm)
            ims[0].callbacksSM.connect('changed', _ImageColourTracker(ims[k]))

        # Add common colorbar
        plt.subplots_adjust(bottom=0.2)
        c_bar_ax = fig.add_axes([0.2, 0.08, 0.6, 0.04])
        c_bar = fig.colorbar(ims[0], c_bar_ax, orientation='horizontal')
        c_bar.solids.set_edgecolor("face")

        if fixed_clim is not None:
            # Force colorbar limits
            c_bar.set_clim(fixed_clim)

        if cbar_label is not None:
            # Set colorbar label
            c_bar.set_label(cbar_label)

    return fig


def mask_img_from_coords(img, coords):
    """
    Mask coordinates in an image.

    The coordinates `coords` in the image `img` are masked such that only the
    cooordinates are shown.

    Parameters
    ----------
    img : ndarray
        The image to mask.
    coords : ndarray
        The coordinates arranged into a 2D array, such that each row is a
        coordinate pair (x, y).

    Returns
    -------
    masked_img : numpy.ma.MaskedArray
        The masked image.

    See Also
    --------
    magni.imaging.measurements : Further description of the coordinate format.

    Examples
    --------
    For example, display only center pixel in a 3-by-3 image

    >>> import numpy as np
    >>> from magni.imaging.visualisation import mask_img_from_coords
    >>> img = np.arange(9).reshape(3, 3)
    >>> coords = np.array([[1, 1]])
    >>> mask_img_from_coords(img, coords)
    masked_array(data =
     [[-- -- --]
     [-- 4 --]
     [-- -- --]],
                 mask =
     [[ True  True  True]
     [ True False  True]
     [ True  True  True]],
           fill_value = 999999)
    <BLANKLINE>

    """

    @_decorate_validation
    def validate_input():
        _numeric('img', ('integer', 'floating', 'complex'), shape=(-1, -1))
        _numeric('coords', 'integer', shape=(-1, 2), range_='[0;inf)')

    validate_input()

    mask = np.ones_like(img, dtype=np.bool_)
    mask[coords[:, 1], coords[:, 0]] = False
    masked_img = np.ma.array(img, mask=mask)

    return masked_img


def shift_mean(x_mod, x_org):
    """
    Shift the mean value of `x_mod` such that it equals the mean of `x_org`.

    Parameters
    ----------
    x_org : ndarray
        The array which hold the "true" mean value.
    x_mod : ndarray
        The modified copy of `x_org` which must have its mean value shifted.

    Returns
    -------
    shifted_x_mod : ndarray
        A copy of `x_mod` with the same mean value as `x_org`.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.visualisation import shift_mean
    >>> x_org = np.arange(4).reshape(2, 2)
    >>> x_mod = np.ones((2, 2))
    >>> print('{:.1f}'.format(x_org.mean()))
    1.5
    >>> print('{:.1f}'.format(x_mod.mean()))
    1.0
    >>> shifted_x_mod = shift_mean(x_mod, x_org)
    >>> print('{:.1f}'.format(shifted_x_mod.mean()))
    1.5
    >>> np.set_printoptions(suppress=True)
    >>> shifted_x_mod
    array([[ 1.5,  1.5],
           [ 1.5,  1.5]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('x_mod', ('integer', 'floating', 'complex'), shape=(-1, -1))
        _numeric('x_org', ('integer', 'floating', 'complex'),
                 shape=x_mod.shape)

    validate_input()

    return x_mod + (x_org.mean() - x_mod.mean())


def stretch_image(img, max_val, min_val=0):
    """
    Stretch image such that pixels values are in [`min_val`, `max_val`].

    Parameters
    ----------
    img : ndarray
        The (float) image that is to be stretched.
    max_val : int or float
        The maximum value in the stretched image.
    min_val : int or float
        The minimum value in the stretched image.

    Returns
    -------
    stretched_img : ndarray
        A stretched copy of the input image.

    Notes
    -----
    The pixel values in the input image are scaled to lie in the interval
    [`min_val`, `max_val`] using a linear stretch.

    Examples
    --------
    For example, stretch an image between 0 and 1

    >>> import numpy as np
    >>> from magni.imaging.visualisation import stretch_image
    >>> img = np.arange(4, dtype=np.float).reshape(2, 2)
    >>> stretched_img = stretch_image(img, 1)
    >>> np.set_printoptions(suppress=True)
    >>> stretched_img
    array([[ 0.        ,  0.33333333],
           [ 0.66666667,  1.        ]])

    or stretch the image between -1 and 1

    >>> stretched_img = stretch_image(img, 1.0, min_val=-1.0)
    >>> stretched_img
    array([[-1.        , -0.33333333],
           [ 0.33333333,  1.        ]])

    or re-stretch the strecthed image between -3.0 and -1.5

    >>> stretched_img = stretch_image(stretched_img, -1.5, min_val=-3.0)
    >>> stretched_img
    array([[-3. , -2.5],
           [-2. , -1.5]])

    or re-stretch that image between 1.25 and 8.00

    >>> stretched_img = stretch_image(stretched_img, 8.00, min_val=1.25)
    >>> stretched_img
    array([[ 1.25,  3.5 ],
           [ 5.75,  8.  ]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('img', 'floating', shape=(-1, -1))
        _numeric('max_val', ('integer', 'floating'))
        _numeric('min_val', ('integer', 'floating'))

        if not max_val > min_val:
            msg = 'max_val ({!r}) must be larger than min_val ({!r})'
            raise ValueError(msg.format(max_val, min_val))

    validate_input()

    min_ = img.min()
    max_ = img.max()

    a = (max_val - min_val) / (max_ - min_)
    b = -a * min_ + min_val

    return a * img + b


def _handle_ticklabels(ax, k, x_ticklabels, y_ticklabels):
    """
    Handle and format ticks and ticklabels for use in imsubplot.

    The `imsubplot` function creates a figure showing an abitrary number of
    subplots. The `imsubplot` function allows for custom ticklabels along the
    x- and y-axes. This function handles the formatting of the ticklabels for a
    given subplot.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The matplotlib axes (subplot) to format ticklabels for.
    k : int
        The axes index, i.e. the subplot number out of the total number of
        subplots.
    x_ticklabels : list or tuple
        The x_ticklabels (as strings or lists of strings) to use for the
        subplots (the default is None, which implies that no x_ticklabels are
        displayed).
    y_ticklabels : list or tuple
        The y_ticklabels (as strings or lists of strings) to use for the
        subplots (the default is None, which implies that no y_ticklabels are
        displayed).

    """

    if x_ticklabels is not None:
        if isinstance(x_ticklabels[0], (list, tuple)):
            x_t = range(len(x_ticklabels[k]))
            x_tl = x_ticklabels[k]
        else:
            # Backwards compatibility
            # Allow for sharing a single set of x_ticklabels across all images.
            x_t = range(len(x_ticklabels))
            x_tl = x_ticklabels

        # Set xticks and xticklabels
        ax.set_xticks(x_t)
        ax.set_xticklabels(x_tl, rotation=90)

    if y_ticklabels is not None:
        if isinstance(y_ticklabels[0], (list, tuple)):
            y_t = range(len(y_ticklabels[k]))
            y_tl = y_ticklabels[k]
        else:
            # Backwards compatibility
            # Allow for sharing a single set of y_ticklabels across all images.
            y_t = range(len(y_ticklabels))
            y_tl = y_ticklabels

        # Set yticks and yticklabels
        ax.set_yticks(y_t)
        ax.set_yticklabels(y_tl)


class _ImageColourTracker():
    """
    Track a common 'clim' in a set of matplotlib image subplots.

    Parameters
    ----------
    tracker : matplotlib.image.AxesImage
        The image instance that must track a given 'clim'.

    """

    def __init__(self, tracker):
        self.tracker = tracker

    def __call__(self, tracked):
        self.tracker.set_clim(tracked.get_clim())



"""
..
    Copyright (c) 2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing public functions for the magni.imaging.measurements
subpackage.

Routine listings
----------------
zigzag_sample_image(h, w, scan_length, num_points, angle=np.pi / 20)
    Function for zigzag sampling an image.
zigzag_sample_surface(l, w, speed, sample_rate, time, angle=np.pi / 20)
    Function for zigzag sampling a surface.

"""

from __future__ import division

import numpy as np

from magni.imaging.measurements import _util
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


__all__ = ['zigzag_sample_image', 'zigzag_sample_surface']

_min_l = _util.min_l
_min_w = _util.min_w
_min_speed = _util.min_speed
_min_sample_rate = _util.min_sample_rate
_min_time = _util.min_time
_min_scan_length = _util.min_scan_length
_min_num_points = _util.min_num_points


def zigzag_sample_image(h, w, scan_length, num_points, angle=np.pi / 20):
    r"""
    Sample an image using a zigzag pattern.

    The coordinates (in units of pixels) resulting from sampling an image of
    size `h` times `w` using a zigzag pattern are determined. The `scan_length`
    determines the length of the path scanned whereas `num_points` indicates
    the number of samples taken on that path.

    Parameters
    ----------
    h : int
        The height of the area to scan in units of pixels.
    w : int
        The width of the area to scan in units of pixels.
    scan_length : float
        The length of the path to scan in units of pixels.
    num_points : int
        The number of samples to take on the scanned path.
    angle : float
        The angle measured in radians by which the lines deviate from being
        horizontal (the default is pi / 20).

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the height `h` is measured along the
    y-axis.

    The `angle` is measured clockwise relative to horizontal and is limited
    to the interval :math:`\left(0;\arctan\left(\frac{h}{w}\right)\right)`.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import zigzag_sample_image
    >>> h = 10
    >>> w = 10
    >>> scan_length = 50.0
    >>> num_points = 12
    >>> np.set_printoptions(suppress=True)
    >>> zigzag_sample_image(h, w, scan_length, num_points)
    array([[ 0.5       ,  0.5       ],
           [ 4.98949246,  1.21106575],
           [ 9.47898491,  1.9221315 ],
           [ 5.03152263,  2.63319725],
           [ 0.54203017,  3.344263  ],
           [ 4.94746229,  4.05532875],
           [ 9.43695474,  4.7663945 ],
           [ 5.0735528 ,  5.47746025],
           [ 0.58406034,  6.188526  ],
           [ 4.90543212,  6.89959175],
           [ 9.39492457,  7.6106575 ],
           [ 5.11558297,  8.32172325]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('h', 'integer', range_='[2;inf)')
        _numeric('w', 'integer', range_='[2;inf)')
        _numeric('scan_length', 'floating',
                 range_='[{};inf)'.format(_min_scan_length))
        _numeric('num_points', 'integer',
                 range_='[{};inf)'.format(_min_num_points))
        _numeric('angle', 'floating', range_='(0;{})'.format(np.arctan(h / w)))

    validate_input()

    coords = zigzag_sample_surface(float(h - 1), float(w - 1), scan_length,
                                   float(num_points - 1), 1., angle)
    coords = coords + 0.5
    return coords


def zigzag_sample_surface(l, w, speed, sample_rate, time, angle=np.pi / 20):
    r"""
    Sample a surface area using a zigzag pattern.

    The coordinates (in units of meters) resulting from sampling an area of
    size `l` times `w` using a zigzag pattern are determined. The scanned path
    is determined from the probe `speed` and the scan `time`.

    Parameters
    ----------
    l : float
        The length of the area to scan in units of meters.
    w : float
        The width of the area to scan in units of meters.
    speed : float
        The probe speed in units of meters/second.
    sample_rate : float
        The sample rate in units of Hertz.
    time : float
        The scan time in units of seconds.
    angle : float
        The angle measured in radians by which the lines deviate from being
        horizontal (the default is pi / 20).

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    Notes
    -----
    The orientation of the coordinate system is such that the width `w` is
    measured along the x-axis whereas the length `l` is measured along the
    y-axis.

    The `angle` is measured clockwise relative to horizontal and is limited
    to the interval :math:`\left(0;\arctan\left(\frac{h}{w}\right)\right)`.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import zigzag_sample_surface
    >>> l = 1e-6
    >>> w = 1e-6
    >>> speed = 7e-7
    >>> sample_rate = 1.0
    >>> time = 12.0
    >>> np.set_printoptions(suppress=True)
    >>> zigzag_sample_surface(l, w, speed, sample_rate, time)
    array([[ 0.        ,  0.        ],
           [ 0.00000069,  0.00000011],
           [ 0.00000062,  0.00000022],
           [ 0.00000007,  0.00000033],
           [ 0.00000077,  0.00000044],
           [ 0.00000054,  0.00000055],
           [ 0.00000015,  0.00000066],
           [ 0.00000084,  0.00000077],
           [ 0.00000047,  0.00000088],
           [ 0.00000022,  0.00000099],
           [ 0.00000091,  0.0000009 ],
           [ 0.00000039,  0.0000008 ],
           [ 0.0000003 ,  0.00000069]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('l', 'floating', range_='[{};inf)'.format(_min_l))
        _numeric('w', 'floating', range_='[{};inf)'.format(_min_w))
        _numeric('speed', 'floating', range_='[{};inf)'.format(_min_speed))
        _numeric('sample_rate', 'floating',
                 range_='[{};inf)'.format(_min_sample_rate))
        _numeric('time', 'floating', range_='[{};inf)'.format(_min_time))
        _numeric('angle', 'floating', range_='(0;{})'.format(np.arctan(l / w)))

    validate_input()

    length = w / np.cos(angle)
    height = length * np.sin(angle)
    number = speed * time / length

    coords = np.zeros((int(np.ceil(number)) + 1, 2))
    coords[1::2, 0] = w
    coords[:, 1] = np.arange(np.ceil(number) + 1) * height
    coords[-1] = (coords[-2] +
                  np.remainder(number, 1) * (coords[-1] - coords[-2]))
    coords = coords.repeat(2, axis=0)

    for i in range(coords.shape[0]):
        if coords[i, 1] < 0:
            coords[i] = (coords[i - 1, 0] + (coords[i - 1, 1] - 0) /
                         height * (coords[i, 0] - coords[i - 1, 0]), 0)
            coords[i + 1:, 1] = -coords[i + 1:, 1]
        elif coords[i, 1] > l:
            coords[i] = (coords[i - 1, 0] + (l - coords[i - 1, 1]) /
                         height * (coords[i, 0] - coords[i - 1, 0]), l)
            coords[i + 1:, 1] = 2 * l - coords[i + 1:, 1]

    return _util.sample_lines(coords, speed, sample_rate, time)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module for testing the Magni documentation build process.

"""

from __future__ import division, print_function
import os
import subprocess
import unittest

import magni


class TestDocBuild(unittest.TestCase):
    """
    Test of correct build of magni documentation using Sphinx

    """

    def setUp(self):
        """
        Change directory to magni/doc and run docapi.

        """

        self.stdout = 'Capture of stdout:\n'
        self.stderr = 'Capture of stderr:\n'
        self.cwd = os.getcwd()

        path, name, ext = magni.utils.split_path(__file__)
        magni_base_path = path.rsplit(os.sep, 3)[0]
        os.chdir(os.path.join(magni_base_path, 'doc'))
        try:
            p_sourceclean = subprocess.Popen(['make', 'sourceclean'],
                                             stdout=subprocess.PIPE,
                                             stderr=subprocess.PIPE)
            p_sourceclean.wait()
            self.stdout += str(p_sourceclean.stdout.read())
            self.stderr += str(p_sourceclean.stderr.read())

            p_htmlclean = subprocess.Popen(['make', 'htmlclean'],
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE)
            p_htmlclean.wait()
            self.stdout += str(p_sourceclean.stdout.read())
            self.stderr += str(p_sourceclean.stderr.read())

            p_docapi = subprocess.Popen(['make', 'docapi'],
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE)
            p_docapi.wait()
            self.stdout += str(p_docapi.stdout.read())
            self.stderr += str(p_docapi.stderr.read())

        except OSError as e:
            os.chdir(self.cwd)
            raise e

    def tearDown(self):
        """
        Run sourceclean and change directory back to tests directory.

        """
        try:
            p_sourceclean = subprocess.Popen(['make', 'sourceclean'],
                                             stdout=subprocess.PIPE,
                                             stderr=subprocess.PIPE)
            p_sourceclean.wait()
            self.stdout += str(p_sourceclean.stdout.read())
            self.stderr += str(p_sourceclean.stderr.read())

        finally:
            os.chdir(self.cwd)

    @unittest.skipIf(os.name == 'nt', 'make is not available')
    def test_html_build(self):
        """
        Test the invocation of "make html" to build Sphinx html documentation.

        """

        p_html = subprocess.Popen(['make', 'SPHINXOPTS=-W', 'html'],
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
        returncode = p_html.wait()
        self.stdout += str(p_html.stdout.read())
        self.stderr += str(p_html.stderr.read())

        error_msg = ('Magni HTML doc build failed.\n\n' +
                     'Stdout:\n{!s}\n\nStderr:\n{!s}'.format(
                         self.stdout, self.stderr))
        self.assertEqual(returncode, 0, msg=error_msg)



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing stop criteria for the Generalised Approximate Message Passing,
GAMP, algorithm.

Routine listings
----------------
ValidatedStopCriterion(magni.utils.validation.types.StopCriterion)
    A base class for validated `magni.cs.reconstruction.gamp` stop criteria.
MSEConvergence(ValidatedStopCriterion)
    A mean square error (MSE) convergence stop criterion.
NormalisedMSEConvergence(ValidatedStopCriterion)
    A normalised mean squaure error (NMSE) convergence stop criterion.
Residual(ValidatedStopCriterion)
    A residual based stop criterion.
ResidualMeasurementsRatio(ValidatedStopCriterion)
    A residual-measurements-ratio based stop criterion.

"""

from __future__ import division

import numpy as np

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_once as _validate_once
from magni.utils.validation.types import StopCriterion as _StopCriterion


class ValidatedStopCriterion(_StopCriterion):
    """
    A base class for validated `magni.cs.reconstruction.gamp` stop criteria.

    Parameters
    ----------
    var : dict
        The stop criterion state variables.

    """

    def __init__(self, var):
        super(ValidatedStopCriterion, self).__init__(var)

        @_decorate_validation
        def validate_input():
            _generic('var', 'mapping')

        validate_input()

    @_validate_once
    def compute(self, var):
        """
        Compute the stop criterion value.

        Parameters
        ----------
        var : dict
            The variables used in computing of the stop criterion value.

        Returns
        -------
        stop : bool
            The indicator of whether or not the stop criterion is satisfied.
        value : float
            The stop criterion value.

        Notes
        -----
        This method honors `magni.utils.validation.enable_allow_validate_once`.

        """

        @_decorate_validation
        def validate_input():
            _generic('var', 'mapping')

        validate_input()


class MSEConvergence(ValidatedStopCriterion):
    """
    A mean square error (MSE) convergence stop criterion.

    Parameters
    ----------
    var : dict
        The stop criterion state variables.

    Notes
    -----
    The following state variables are used in this stop criterion:

    * A
    * tolerance

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.cs.reconstruction.gamp.stop_criterion import MSEConvergence
    >>> state = {'tolerance': 1e-3, 'A': np.ones((10,10))}
    >>> variables = {'alpha_bar_prev': np.ones((10, 1)),
    ... 'alpha_bar': np.arange(10).reshape(10, 1)}
    >>> MSE = MSEConvergence(state)
    >>> stop, val = MSE.compute(variables)
    >>> stop
    False
    >>> np.round(val, 2)
    20.5

    """

    def __init__(self, var):
        super(MSEConvergence, self).__init__(var)
        self.n = var['A'].shape[1]
        self.tolerance = var['tolerance']

    def compute(self, var):
        """
        Compute the MSE convergence stop criterion value.

        The GAMP algorithm should converge to a fixed point. This criterion
        is based on the mean squared error of the difference between the
        proposed solution in this iteration and the proposed solution in the
        previous solution.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in calculating of the stop criterion.

        Returns
        -------
        stop : bool
            The indicator of whether or not the stop criterion is satisfied.
        value : float
            The stop criterion value.

        """

        super(MSEConvergence, self).compute(var)

        mse = 1/self.n * np.linalg.norm(
            var['alpha_bar_prev'] - var['alpha_bar'])**2
        stop = mse < self.tolerance

        return stop, mse


class NormalisedMSEConvergence(ValidatedStopCriterion):
    """
    A normalised mean squaure error (NMSE) convergence stop criterion.

    Parameters
    ----------
    var : dict
        The stop criterion state variables.

    Notes
    -----
    The following state variables are used in this stop criterion:

    * tolerance

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.cs.reconstruction.gamp.stop_criterion import (
    ... NormalisedMSEConvergence)
    >>> state = {'tolerance': 1e-3}
    >>> variables = {'alpha_bar_prev': np.ones((10, 1)),
    ... 'alpha_bar': np.arange(10).reshape(10, 1)}
    >>> NMSE = NormalisedMSEConvergence(state)
    >>> stop, val = NMSE.compute(variables)
    >>> stop
    False
    >>> np.round(val, 2)
    20.5

    """

    def __init__(self, var):
        super(NormalisedMSEConvergence, self).__init__(var)
        self.tolerance = var['tolerance']

    def compute(self, var):
        """
        Compute the normalised MSE convergence stop criterion value.

        The GAMP algorithm should converge to a fixed point. This criterion
        is based on the mean squared error of the difference between the
        proposed solution in this iteration and the proposed solution in the
        previous solution normalised by the mean squared error of the proposed
        solution in the previous iteration.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in calculating of the stop criterion.

        Returns
        -------
        stop : bool
            The indicator of whether or not the stop criterion is satisfied.
        value : float
            The stop criterion value.

        """

        super(NormalisedMSEConvergence, self).compute(var)

        se = np.linalg.norm(var['alpha_bar_prev'] - var['alpha_bar'])**2
        norm = np.linalg.norm(var['alpha_bar_prev'])**2
        if norm > 0:
            value = se / norm
            stop = se < self.tolerance * norm
        else:
            # Previous solution was zero-vector (most likely first iteration)
            value = se
            stop = False

        return stop, value


class Residual(ValidatedStopCriterion):
    """
    A residual based stop criterion.

    Parameters
    ----------
    var : dict
        The stop criterion state variables.

    Notes
    -----
    The following state variables are used in this stop criterion:

    * y
    * tolerance
    * A

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.cs.reconstruction.gamp.stop_criterion import Residual
    >>> state = {'tolerance': 1e-3, 'y': np.ones((10, 1)),
    ... 'A': np.ones((10,10))}
    >>> variables = {'A_dot_alpha_bar': np.arange(10).reshape(10, 1)}
    >>> Res = Residual(state)
    >>> stop, val = Res.compute(variables)
    >>> stop
    False
    >>> np.round(val, 2)
    20.5

    """

    def __init__(self, var):
        super(Residual, self).__init__(var)
        self.y = var['y']
        self.tolerance = var['tolerance']
        self.m = var['A'].shape[0]

    def compute(self, var):
        """
        Compute the residual stop criterion value.

        If the noise level is (approximately) known, the GAMP iterations may be
        stopped once the residual is on the order of the noise level. This
        stopping criterion is based on the mean sqaured error of the residual.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in calculating of the stop criterion.

        Returns
        -------
        stop : bool
            The indicator of whether or not the stop criterion is satisfied.
        value : float
            The stop criterion value.

        """

        super(Residual, self).compute(var)

        r_mse = 1/self.m * np.linalg.norm(self.y - var['A_dot_alpha_bar'])**2
        stop = r_mse < self.tolerance

        return stop, r_mse


class ResidualMeasurementsRatio(ValidatedStopCriterion):
    """
    A residual-measurements-ratio based stop criterion.

    Parameters
    ----------
    var : dict
        The stop criterion state variables.

    Notes
    -----
    The following state variables are used in this stop criterion:

    * y
    * tolerance

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.cs.reconstruction.gamp.stop_criterion import (
    ... ResidualMeasurementsRatio)
    >>> state = {'tolerance': 1e-3, 'y': np.ones((10, 1))}
    >>> variables = {'A_dot_alpha_bar': np.arange(10).reshape(10, 1)}
    >>> ResMeasRat = ResidualMeasurementsRatio(state)
    >>> stop, val = ResMeasRat.compute(variables)
    >>> stop
    False
    >>> np.round(val, 2)
    14.32

    """

    def __init__(self, var):
        super(ResidualMeasurementsRatio, self).__init__(var)
        self.y = var['y']
        self.tolerance = var['tolerance']

    def compute(self, var):
        """
        Compute the residual-measurements-ratio stop criterion value.

        If the noise level is (approximately) known, the GAMP iterations may be
        stopped once the residual is on the order of the noise level. This
        stopping criterion is based on ratio of the mean sqaured error of the
        residual to the mean squared error of the measurements.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in calculating of the stop criterion.

        Returns
        -------
        stop : bool
            The indicator of whether or not the stop criterion is satisfied.
        value : float
            The stop criterion value.

        """

        super(ResidualMeasurementsRatio, self).compute(var)

        r_norm = np.linalg.norm(self.y - var['A_dot_alpha_bar'])
        stop = r_norm < self.tolerance * np.linalg.norm(self.y)

        return stop, r_norm



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality for aiding in quest for more reproducible
research.

Routine listings
----------------
data
    Module providing functions that return various data about the system,
    magni, files, etc.
io
    Module providing input/output functions to databases containing results
    from reproducible research.

"""

from magni.reproducibility import data
from magni.reproducibility import io



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing a robust configger class.

Routine listings
----------------
Configger(object)
    Provide functionality to access a set of configuration options.

Notes
-----
This module does not itself contain any configuration options and thus has no
access to any configuration options unlike the other config modules of `magni`.

"""

from __future__ import division
from itertools import chain

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric


class Configger(object):
    """
    Provide functionality to access a set of configuration options.

    The set of configuration options, their default values, and their
    validation schemes are specified upon initialisation.

    Parameters
    ----------
    params : dict
        The configuration options and their default values.
    valids : dict
        The validation schemes of the configuration options.

    See Also
    --------
    magni.utils.validation : Validation.

    Notes
    -----
    `valids` must contain the same keys as `params`. For each key in 'valids',
    the first value is the validation function ('generic', 'levels', or
    'numeric'), whereas the remaining values are passed to that validation
    function.

    Examples
    --------
    Instantiate Configger with the parameter 'key' with default value 'default'
    which can only assume string values.

    >>> import magni
    >>> from magni.utils.config import Configger
    >>> valid = magni.utils.validation.validate_generic(None, 'string')
    >>> config = Configger({'key': 'default'}, {'key': valid})

    The number of parameters can be retrieved as the length:

    >>> len(config)
    1

    That parameter can be retrieved in a number of ways:

    >>> config['key']
    'default'

    >>> for key, value in config.items():
    ...     print('key: {!r}, value: {!r}'.format(key, value))
    key: 'key', value: 'default'

    >>> for key in config.keys():
    ...     print('key: {!r}'.format(key))
    key: 'key'

    >>> for value in config.values():
    ...     print('value: {!r}'.format(value))
    value: 'default'

    Likewise, the parameter can be changed in a number of ways:

    >>> config['key'] = 'value'
    >>> config['key']
    'value'

    >>> config.update({'key': 'value changed by dict'})
    >>> config['key']
    'value changed by dict'

    >>> config.update(key='value changed by keyword')
    >>> config['key']
    'value changed by keyword'

    Finally, the parameter can be reset to the default value at any point:

    >>> config.reset()
    >>> config['key']
    'default'

    """

    _funcs = {'generic': _generic, 'levels': _levels, 'numeric': _numeric}

    def __init__(self, params, valids):
        @_decorate_validation
        def validate_input():
            _generic('params', 'mapping')
            _levels('valids', (
                _generic(None, 'mapping', has_keys=tuple(params.keys())),
                _generic(None, 'explicit collection')))

            for key in valids.keys():
                _generic(('valids', key, 0), 'string',
                         value_in=tuple(self._funcs.keys()))

        validate_input()

        self._default = params.copy()
        self._params = params
        self._valids = valids

    def __getitem__(self, name):
        """
        Get the value of a configuration parameter.

        Parameters
        ----------
        name : str
            The name of the parameter.

        Returns
        -------
        value : None
            The value of the parameter.

        """

        @_decorate_validation
        def validate_input():
            _generic('name', 'string', value_in=tuple(self._params.keys()))

        validate_input()

        return self._params[name]

    def __len__(self):
        """
        Get the number of configuration parameters.

        Returns
        -------
        length : int
            The number of parameters.

        """

        return len(self._params)

    def __setitem__(self, name, value):
        """
        Set the value of a configuration parameter.

        The value is validated according to the validation scheme of that
        parameter.

        Parameters
        ----------
        name : str
            The name of the parameter.
        value : None
            The new value of the parameter.

        """

        @_decorate_validation
        def validate_input():
            _generic('name', 'string', value_in=tuple(self._params.keys()))
            validation = self._valids[name]
            self._funcs[validation[0]]('value', *validation[1:])

        validate_input()

        self._params[name] = value

    def get(self, key=None):
        """
        Deprecated method.

        See Also
        --------
        Configger.__getitem__ : Replacing method.
        Configger.items : Replacing method.
        Configger.keys : Replacing method.
        Configger.values : Replacing method.

        """

        raise DeprecationWarning("'get' will be removed in version 1.3.0 - "
                                 "use 'var[name]', 'items', 'keys', or "
                                 "'values' instead.")

        if key is None:
            return dict(self.items())
        else:
            return self[key]

    def items(self):
        """
        Get the configuration parameters as key, value pairs.

        Returns
        -------
        items : set-like
            The list of parameters.

        """

        for key in self.keys():
            yield (key, self[key])

    def keys(self):
        """
        Get the configuration parameter keys.

        Returns
        -------
        keys : set-like
            The keys.

        """

        return self._params.keys()

    def reset(self):
        """
        Reset the parameter values to the default values.

        """

        self._params = self._default.copy()

    def set(self, dictionary={}, **kwargs):
        """
        Deprecated method.

        See Also
        --------
        Configger.__setitem__ : Replacing function.

        """

        raise DeprecationWarning("'set' will be removed in version 1.3.0 - "
                                 "use 'var[name] = value' or 'update' "
                                 "instead.")

        self.update(dictionary, **kwargs)

    def update(self, params={}, **kwargs):
        """
        Update the value of one or more configuration parameters.

        Each value is validated according to the validation scheme of that
        parameter.

        Parameters
        ----------
        params : dict, optional
            A dictionary containing the key and values to update. (the default
            value is an empty dictionary)
        kwargs : dict
            Keyword arguments being the key and values to update.

        """

        @_decorate_validation
        def validate_input():
            _generic('params', 'mapping', keys_in=tuple(self._params.keys()))

            for name, var in (('params', params), ('kwargs', kwargs)):
                for key in var.keys():
                    validation = self._valids[key]
                    self._funcs[validation[0]]((name, key), *validation[1:])

        validate_input()

        if params is not None:
            for key, value in params.items():
                self[key] = value

        if len(kwargs) > 0:
            for key, value in kwargs.items():
                self[key] = value

    def values(self):
        """
        Get the configuration parameter values.

        Returns
        -------
        values : set-like
            The values.

        """

        for key in self.keys():
            yield self[key]



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality for disabling validation.

The disabling functionality is provided through a decorator for validation
functions and a function for disabling validation. Furthermore, the present
module provides functionality which is internal to `magni.utils.validation`.

Routine listings
----------------
decorate_validation(func)
    Decorate a validation function to allow disabling of validation checks.
disable_validation()
    Disable validation checks in `magni`.
enable_validate_once()
    Enable validating inputs only once in certain functions in `magni`.
get_var(name)
    Retrieve the value of a variable through call stack inspection.
report(type, description, format_args=(), var_name=None, var_value=None,
    expr='{}', prepend='')
    Raise an exception.
validate_once(func)
    Decorate a function to allow for a one-time input validation only.

"""

from __future__ import division

import functools
import inspect
import os


_disabled = False
_validate_once_enabled = False


def decorate_validation(func):
    """
    Decorate a validation function to allow disabling of validation checks.

    Parameters
    ----------
    func : function
        The validation function to be decorated.

    Returns
    -------
    func : function
        The decorated validation function.

    See Also
    --------
    disable_validation : Disabling of validation checks.

    Notes
    -----
    This decorater wraps the validation function in another function which
    checks if validation has been disabled. If validation has been disabled,
    the validation function is not called. Otherwise, the validation function
    is called.

    Examples
    --------
    See `disable_validation` for an example.

    """

    @functools.wraps(func)
    def wrapper():
        if not _disabled:
            func()

    return wrapper


def disable_validation():
    """
    Disable validation checks in `magni`.

    See Also
    --------
    decorate_validation : Decoration of validation functions.

    Notes
    -----
    This function merely sets a global flag and relies on `decorate_validation`
    to perform the actual disabling.

    Examples
    --------
    An example of a function which accepts only an integer as argument:

    >>> import magni
    >>> def test(arg):
    ...     @magni.utils.validation.decorate_validation
    ...     def validate_input():
    ...         magni.utils.validation.validate_numeric('arg', 'integer')
    ...     validate_input()

    If the function is called with anything but an integer, it fails:

    >>> try:
    ...     test('string')
    ... except BaseException:
    ...     print('An exception occured')
    ... else:
    ...     print('No exception occured')
    An exception occured

    However, if validation is disabled, the same call does not fail:

    >>> from magni.utils.validation import disable_validation
    >>> disable_validation()
    >>> try:
    ...     test('string')
    ... except BaseException:
    ...     print('An exception occured')
    ... else:
    ...     print('No exception occured')
    No exception occured

    Although strongly discouraged, validation can be re-enabled after being
    disabled in the following way:

    >>> import magni.utils.validation._util
    >>> magni.utils.validation._util._disabled = False

    As this interface is not public it may be subject to changes in future
    releases without further notice.

    """

    global _disabled
    _disabled = True


def enable_validate_once():
    """
    Enable validating inputs only once in certain functions in `magni`.

    See Also
    --------
    validate_once : Decoration of functions.

    Notes
    -----
    This function merely sets a global flag and relies on `validate_once`
    to implement the actual "validate once" functionality.

    Examples
    --------
    An example of a function which accepts only an integer as argument:

    >>> import magni
    >>> @magni.utils.validation.validate_once
    ... def test(arg):
    ...     @magni.utils.validation.decorate_validation
    ...     def validate_input():
    ...         magni.utils.validation.validate_numeric('arg', 'integer')
    ...     validate_input()

    If "validate once" is enabled, the function validation is only called on
    its first run:

    >>> magni.utils.validation.enable_validate_once()
    >>> try:
    ...     test('string')
    ... except BaseException:
    ...     print('An exception occured')
    ... else:
    ...     print('No exception occured')
    An exception occured

    >>> try:
    ...     test('string')
    ... except BaseException:
    ...     print('An exception occured')
    ... else:
    ...     print('No exception occured')
    No exception occured

    """

    global _validate_once_enabled
    _validate_once_enabled = True


def get_var(name):
    """
    Retrieve the value of a variable through call stack inspection.

    `name` must refer to a variable in the parent scope of the function or
    method decorated by `magni.utils.validation.decorate_validation` which is
    closest to the top of the call stack. If `name` is a string then there must
    be a variable of that name in that scope. If `name` is a set-like object
    then there must be a variable having the first value in that set-like
    object as name. The remaining values are used as keys/indices on the
    variable to obtain the variable to be validated. For example, the `name`
    ('name', 0, 'key') refers to the variable "name[0]['key']".

    Parameters
    ----------
    name : None
        The name of the variable to be retrieved.

    Returns
    -------
    var : None
        The value of the retrieved variable.

    Notes
    -----
    The present function searches the call stack from top to bottom until it
    finds a function named 'wrapper' defined in this file. That is, until it
    finds a decorated validation function. The present function then looks up
    the variable indicated by `name` in the parent scope of that decorated
    validation function.

    """

    frame = inspect.currentframe()
    path = frame.f_code.co_filename
    index = path.rindex(os.path.sep)
    path = path[:index]
    frame = frame.f_back.f_back
    code = frame.f_code

    try:
        while code.co_firstlineno != 73 or code.co_filename[:index] != path:
            frame = frame.f_back
            code = frame.f_code
    except AttributeError:
        report(RuntimeError, 'Validation functions must be decorated.',
               prepend='Invalid validation call: ')

    if not isinstance(name, (list, tuple)):
        name = (name,)

    lookups = name[1:]
    name = name[0]

    try:
        var = frame.f_back.f_locals[name]
    except KeyError:
        report(NameError, 'must refer to an argument of {!r}.',
               frame.f_back.f_code.co_name, var_name='name', var_value=name,
               prepend='Invalid validation call: ')

    try:
        for i, lookup in enumerate(lookups):
            var = var[lookup]
    except Exception:
        report(LookupError, 'must have the key or index, {!r}.', lookup,
               var_name=[name] + list(lookups[:i]))

    return var


def report(type_, description, format_args=(), var_name=None, var_value=None,
           expr='{}', prepend=''):
    """
    Raise an exception.

    The type of the exception is given by `type_`, and the message can take on
    many forms. This ranges from a single description to a description
    formatted using a number of arguments, prepended by an expression using a
    variable name followed by the variable value, prepended by another
    description.

    Parameters
    ----------
    type_ : type
        The exception type.
    description : str
        The core description.
    format_args : list or tuple
        The arguments which `description` is formatted with. (the default is
        (), which implies no arguments)
    var_name : None
        The name of the variable which the description concerns. (the default
        is None, which implies that no variable name and value is prepended to
        the description)
    var_value : None
        The value of the variable which the description concerns. (the default
        is None, which implies that the value is looked up from the `var_name`
        in the call stack if `var_name` is not None)
    expr : str
        The expression to evaluate with the variable name. (the default is
        '{}', which implies that the variable value is used directly)
    prepend : str
        The text to prepend to the message generated by the previous arguments.
        (the default is '')

    Notes
    -----
    `name` must refer to a variable in the parent scope of the function or
    method decorated by `magni.utils.validation.decorate_validation` which is
    closest to the top of the call stack. If `name` is a string then there must
    be a variable of that name in that scope. If `name` is a set-like object
    then there must be a variable having the first value in that set-like
    object as name. The remaining values are used as keys/indices on the
    variable to obtain the variable to be validated. For example, the `name`
    ('name', 0, 'key') refers to the variable "name[0]['key']".

    """

    try:
        descr = description.format(format_args)
    except IndexError:
        descr = description.format(*format_args)

    if var_name is not None:
        value = get_var(var_name) if var_value is None else var_value
        value = eval(expr.replace('{}', 'value'))
        descr = ', {!r}, '.format(value) + descr

        if not isinstance(var_name, (list, tuple)):
            var_name = (var_name,)

        path = ['[{!r}]'.format(part) for part in var_name[1:]]
        path = var_name[0] + ''.join(path)
        descr = ('>>' + expr + '<<').format(path) + descr

        descr = 'The value(s) of ' + descr

    descr = prepend + descr
    raise type_(descr)


def validate_once(func):
    """
    Decorate a function to allow for a one-time input validation only.

    Parameters
    ----------
    func : function
        The validation function to be decorated.

    Returns
    -------
    func : function
        The decorated validation function.

    See Also
    --------
    enable_validate_once : Enabling allow validate once functionality.

    Notes
    -----
    This decorater wraps any function. It checks if the "validate once" has
    been enabled. If it has been enabled, the validation function is only
    called on the first run. Otherwise, the validation function is always
    called if not otherwise disabled.

    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        global _disabled

        if _disabled or not _validate_once_enabled:
            return func(*args, **kwargs)
        elif not wrapper.has_run:
            wrapper.has_run = True
            return func(*args, **kwargs)
        else:
            try:
                _disabled = True
                return func(*args, **kwargs)
            finally:
                _disabled = False

    wrapper.has_run = False

    return wrapper



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing problem suite instance and noise generation functionality.

The problem suite instances consist of a matrix, A, and a coefficient vector,
alpha, with which the measurement vector, y, can be generated (with or without
noise from the noise vector e)

Routine listings
----------------
generate_matrix(m, n)
    Generate a matrix belonging to a specific problem suite.
generate_noise(m, n, k)
    Generate a noise vector of a specific type.
generate_vector(n, k)
    Generate a vector belonging to a specific problem suite.

See also
--------
magni.cs.phase_transition._config: Configuration options.

Notes
-----
The matrices and vectors generated in this module use the numpy.random
submodule. Consequently, the calling script or function should control the seed
to ensure reproducibility.

The choice of non-zero indices in the coefficient vector is controlled by the
configuration option 'support_structure' whereas the distribution of the
non-zero coefficients is controlled by the configuration option 'coefficient'.

Examples
--------
For example generate a sample from the USE/Rademacher problem suite:

>>> import numpy as np, magni
>>> from magni.cs.phase_transition import _data
>>> m, n, k = 400, 800, 100
>>> A = _data.generate_matrix(m, n)
>>> alpha = _data.generate_vector(n, k)
>>> y = np.dot(A, alpha)

Or generate a problem suite instance with "linear" support distribution.

>>> support_distrib = np.reshape(np.arange(n, dtype=np.float) + 1, (n, 1))
>>> support_distrib /= np.sum(support_distrib)
>>> magni.cs.phase_transition.config['support_distribution'] = support_distrib
>>> A = _data.generate_matrix(m, n)
>>> alpha = _data.generate_vector(n, k)
>>> y = np.dot(A, alpha)

Or generate an AWGN noise vector based on a 40 dB SNR

>>> magni.cs.phase_transition.config['noise'] = 'AWGN'
>>> e = _data.generate_noise(m, n, k)

"""

from __future__ import division

import numpy as np

import magni.imaging as _imaging
from magni.cs.phase_transition import config as _conf
from magni.utils.matrices import MatrixCollection as _MatrixCollection
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric


def generate_matrix(m, n):
    """
    Generate a matrix belonging to a specific problem suite.

    The available matrices are

    * A random matrix drawn from the Uniform Spherical Ensemble (USE).
    * A fixed uniformly row sub-sampled DCT matrix ensemble (RandomDCT2D).
    * An option to use custom matrix factory (see notes below).

    See Notes for a description of these matrices. Which of the available
    matrices is used, is specified as a configuration option.

    Parameters
    ----------
    m : int
        The number of rows.
    n : int
        The number of columns.

    Returns
    -------
    A : ndarray
        The generated matrix.

    See Also
    --------
    magni.cs.phase_transition.config : Configuration options.
    magni.utils.matrices.MatrixCollection : Fast transform implementation.

    Notes
    -----
    The Uniform Spherical Ensemble:
        The matrices of this ensemble have i.i.d. Gaussian entries of mean zero
        and variance one. Its columns are then normalised to have unit length.
    Fixed uniformly row sub-sampled DCT ensemble:
        The matrices of this ensemble correspond to the combination of a 2D
        array sub-sampled using a uniform point pattern and a 2D Discrete
        Cosine Transform (DCT) matrix. The combined matrix is implemented as a
        fast transform with a DCT based on an FFT routine.
    Custom matrix factory:
        The matrix generation is delegated to the configured
        `custom_system_matrix_factory` callable which is expected to take the
        arguments `m`, `n` and return `A`.

    """

    @_decorate_validation
    def validate_input():
        _numeric('n', 'integer', range_='[1;inf)')
        _numeric('m', 'integer', range_='[1;{}]'.format(n))

    @_decorate_validation
    def validate_output():
        _numeric('A', ('integer', 'floating', 'complex'), shape=(m, n))

    validate_input()

    system_matrix = _conf['system_matrix']

    if system_matrix == 'USE':
        A = np.float64(np.random.randn(m, n))
        A = A / np.linalg.norm(A, axis=0)

    elif system_matrix == 'RandomDCT2D':
        n_sqrt = np.sqrt(n)
        if not n_sqrt.is_integer():
            raise ValueError(
                'When using a RandomDCT2D system matrix, ' +
                'the value of >>{!r}<<, {!r} must be a square number'.format(
                    'n', n))

        n_sqrt = int(n_sqrt)

        # Random pixel subsampling operator
        points = np.random.choice(np.arange(n), size=m, replace=False)
        points.sort()
        coords = np.vstack([points // n_sqrt, points % n_sqrt]).T
        Phi = _imaging.measurements.construct_measurement_matrix(
            coords, n_sqrt, n_sqrt)
        assert Phi.shape[0] == m

        # DCT dictionary
        Psi = _imaging.dictionaries.get_DCT((n_sqrt, n_sqrt))

        A = _MatrixCollection((Phi, Psi))

    elif system_matrix == 'custom':
        A = _conf['custom_system_matrix_factory'](m, n)

    validate_output()

    return A


def generate_noise(m, n, k):
    """
    Generate a noise vector of a specific type.

    The available types are:

    * AWGN : Additive White Gaussian Noise
    * AWLN : Additive White Laplacian Noise
    * custom : The noise generation is delegated to the configured
      `custom_noise_factory` callable which is expected to take the arguments
      `m`, `n`, `k`, `noise_power`.

    Which of the available types is used, is specified as a configuration
    option.

    Parameters
    ----------
    m : int
        The number of rows.
    n : int
        The number of columns.
    k : int
        The number of non-zero coefficients.

    Returns
    -------
    e : ndarray
        The generated noise vector.

    See Also
    --------
    magni.cs.phase_transition.config : Configuration options.

    Notes
    -----
    The noise power is computed from the configured `SNR` and the theoretical
    ensemble variance of the coefficients generated by `generate_cofficients`.

    """

    @_decorate_validation
    def validate_input():
        _numeric('n', 'integer', range_='[1;inf)')
        _numeric('m', 'integer', range_='[1;{}]'.format(n))
        _numeric('k', 'integer', range_='[1;{}]'.format(n))

    @_decorate_validation
    def validate_output():
        _numeric('e', ('integer', 'floating', 'complex'), shape=(m, 1))

    validate_input()

    coefficients = _conf['coefficients']
    noise_type = _conf['noise']
    SNR = _conf['SNR']
    if noise_type is None:
        raise RuntimeError('The noise type has not been configured.')

    tau = k / n
    size = (m, 1)

    noise_generators = {
        'AWGN': lambda s: np.random.normal(scale=np.sqrt(s), size=size),
        'AWLN': lambda s: np.random.laplace(scale=np.sqrt(s / 2), size=size),
        'custom': lambda s: _conf['custom_noise_factory'](m, n, k, s)
    }

    # Generally the signal power is the coefficient variance scaled by the
    # signal density tau (number of non-zero coefficients).
    # This dictionary must be synced with the coeffient generators dictionary
    # in the generate_vector function below.
    signal_powers = {
        'rademacher': 1.0 * tau,
        'gaussian': 1.0 * tau,
        'laplace': 2.0 * tau,
        'bernoulli': 1.0 * tau
        }

    noise_power = 10**(-SNR/10) * signal_powers[coefficients]
    e = noise_generators[noise_type](noise_power)

    validate_output()

    return e


def generate_vector(n, k):
    """
    Generate a vector belonging to a specific problem suite.

    The available ensembles are:

    * Gaussian
    * Rademacher
    * Laplace
    * Bernoulli

    See Notes for a description of the ensembles. Which of the available
    ensembles is used, is specified as a configuration option. Note, that the
    non-zero `k` non-zero coefficients are the `k` first entries if no support
    structure specified in the configuration.

    Parameters
    ----------
    n : int
        The length of the vector.
    k : int
        The number of non-zero coefficients.

    Returns
    -------
    alpha : ndarray
        The generated vector.

    See Also
    --------
    magni.cs.phase_transition.config : Configuration options.

    Notes
    -----
    The Gaussian ensemble:
        The non-zero coefficients are drawn from the normal Gaussian
        distribution.
    The Rademacher ensemble:
        The non-zero coefficients are drawn from the constant amplitude with
        random signs ensemble.
    The Laplace ensemble:
        The non-zero coefficients are drawn from the zero-mean, unit scale
        Laplace distribution (variance = 2).
    The Bernoulli ensemble:
        The non-zero coefficients are all equal to one.

    """

    @_decorate_validation
    def validate_input():
        _numeric('n', 'integer', range_='[1;inf)')
        _numeric('k', 'integer', range_='[1;{}]'.format(n))

    @_decorate_validation
    def validate_output():
        _numeric('alpha', ('integer', 'floating', 'complex'), shape=(n, 1))

    validate_input()

    alpha = np.zeros((n, 1))
    coefficients = _conf['coefficients']
    support_distribution = _conf['support_distribution']

    # This dictionary must be synced with the signal powers dictionary in the
    # generate_noise function above.
    coefficient_generators = {
        'rademacher': lambda k: np.random.randint(0, 2, k) * 2 - 1,
        'gaussian': lambda k: np.random.randn(k),
        'laplace': lambda k: np.random.laplace(size=k),  # Variance = 2
        'bernoulli': lambda k: np.ones(k)}

    if support_distribution is not None:
        ix = np.random.choice(np.arange(n), size=k, replace=False,
                              p=support_distribution.reshape(-1))
        alpha[ix, 0] = coefficient_generators[coefficients](k)

    else:
        alpha[:k, 0] = coefficient_generators[coefficients](k)

    validate_output()

    return alpha



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the public functionality of the magni.afm.io subpackage.

"""

from collections import defaultdict as _defaultdict
import os

from magni.afm.io._data_attachment import attach_data as _attach_data
from magni.afm.io._object_building import build_object as _build_object
from magni.afm.io._stream_conversion import convert_stream as _convert_stream
from magni.afm.types.image import Buffer as _ImageBuffer
from magni.afm.types.image import Image as _Image
from magni.afm.types.spectroscopy import Buffer as _SpectroscopyBuffer
from magni.afm.types.spectroscopy import Chunk as _SpectroscopyChunk
from magni.afm.types.spectroscopy import Grid as _SpectroscopyGrid
from magni.afm.types.spectroscopy import Point as _SpectroscopyPoint
from magni.afm.types.spectroscopy import Spectroscopy as _Spectroscopy
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic


def read_mi_file(path):
    """
    Read an .mi file given by a path.

    The supported .mi file types are Image and Spectroscopy.

    Parameters
    ----------
    path : str
        The path of the .mi file to read.

    Returns
    -------
    file_ : magni.afm.types.File
        The read .mi file.

    See Also
    --------
    magni.afm.io._stream_conversion : First step of the reading.
    magni.afm.io._object_building : Second step of the reading.
    magni.afm.io._data_attachment : Third step of the reading.

    Notes
    -----
    Depending on the .mi file type, the returned value will be an instance of a
    subclass of `magni.afm.types.File`: `magni.afm.types.image.Image` or
    `magni.afm.types.spectroscopy.Spectroscopy`.

    The reading of an .mi file is logically separated into four steps:

    1. Converting the file stream to flat, basic variables.
    2. Converting the header parameters to a hierarchical object-structure
       mimicking that of `magni.afm.types`.
    3. Attaching the actual data to the hierarchical object-structure.
    4. Instantiating the classes in `magni.afm.types` from the hierarchical
       object structure.

    The functionality needed for each step is grouped in separate modules with
    the functionality needed for the fourth step being grouped in the this
    module.

    Examples
    --------
    An example of how to use this function to read the example .mi file
    provided with the package:

    >>> import os, magni
    >>> from magni.afm.io import read_mi_file
    >>> path = magni.utils.split_path(magni.__path__[0])[0]
    >>> path = path + 'examples' + os.sep + 'example.mi'
    >>> if os.path.isfile(path):
    ...     mi_file = read_mi_file(path)

    """

    @_decorate_validation
    def validate_input():
        _generic('path', 'string')

    validate_input()

    with open(path, 'rb') as f:
        stream = f.read()

    params, data = _convert_stream(stream)
    obj = _build_object(params)
    _attach_data(obj, data)

    if obj['attrs']['fileType'].lower() == 'image':
        for i, buffer_ in enumerate(obj['buffers']):
            obj['buffers'][i] = _ImageBuffer(buffer_['attrs'], buffer_['data'])

        obj = _Image(obj['attrs'], obj['buffers'])
    else:  # 'spectroscopy'
        obj = _instantiate_spectroscopy(obj)

    return obj


def _instantiate_spectroscopy(obj):
    """
    Instantiate a Spectroscopy object from a hierarchical object-structure.

    Parameters
    ----------
    obj : object
        The hierarchical object-structure.

    Returns
    -------
    spectroscopy : magni.afm.types.spectroscopy.Spectroscopy
        The instantiated Spectroscopy object.

    See Also
    --------
    read_mi_file : Function using the present function.

    Notes
    -----
    This function instantiates the sweep buffer and the spectroscopy file.

    """

    items = []

    for item in obj['buffers'][0]['data']:
        if item['type'] == 'grid':
            points = [[_SpectroscopyPoint(point['attrs'], []) for point in row]
                      for row in item['points']]
            items.append(_SpectroscopyGrid(item['attrs'], points))
        else:  # 'point'
            items.append(_SpectroscopyPoint(item['attrs'], []))

    sweep_buffer = _SpectroscopyBuffer(obj['buffers'][0]['attrs'], items)
    data_buffers = [_instantiate_spectroscopy_buffer(buffer_, sweep_buffer)
                    for buffer_ in obj['buffers'][1:]]

    return _Spectroscopy(obj['attrs'], [sweep_buffer] + data_buffers)


def _instantiate_spectroscopy_buffer(data_buffer, sweep_buffer):
    """
    Instantiate a spectroscopy Buffer object from a hierarchical
    object-structure.

    Parameters
    ----------
    obj : object
        The hierarchical object-structure.

    Returns
    -------
    buffer_ : magni.afm.types.spectroscopy.Buffer
        The instantiated spectroscopy Buffer.

    See Also
    --------
    _instantiate_spectroscopy : Function using the present function.

    """

    chunks = _defaultdict(lambda: [])

    for chunk in data_buffer['data']:
        key = chunk['attrs'].get('pointIndex')
        chunks[key].append(_SpectroscopyChunk(chunk['attrs'], chunk['data']))

    if len(sweep_buffer.data) == 0:
        chunks[None].extend(chunks[0])

    items = []

    for item in sweep_buffer.data:
        if isinstance(item, _SpectroscopyGrid):
            points = [[_SpectroscopyPoint(point.attrs,
                                          chunks[point.attrs['index']])
                       for point in row] for row in item.points]
            items.append(_SpectroscopyGrid(item.attrs, points))
        elif isinstance(item, _SpectroscopyPoint):
            items.append(_SpectroscopyPoint(item.attrs,
                                            chunks[item.attrs['index']]))

    return _SpectroscopyBuffer(data_buffer['attrs'], items + chunks[None])



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing public functions for the magni.imaging.measurements
subpackage.

Routine listings
----------------
construct_pixel_mask(h, w, pixels)
    Construct a binary pixel mask.
unique_pixels(coords)
    Function for determining unique pixels from a set of coordinates.

"""

from __future__ import division

import numpy as np

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


__all__ = ['construct_pixel_mask', 'unique_pixels']

# In principle most of the AFM-scanning related parameters should just be
# positive, however we have settled for:
min_l = 1e-9  # [m]
min_w = 1e-9  # [m]
min_speed = 1e-9  # [m/s]
min_sample_rate = 1.0  # [Hz]
min_time = 1.0  # [s]
min_scan_length = 1e-9  # [m]
min_num_points = 1  # []


def construct_pixel_mask(h, w, pixels):
    """
    Construct a binary pixel mask.

    An image (2D array) of shape `w` x `h` is created where all `pixels` are
    marked True.

    Parameters
    ----------
    h : int
        The height of the image in pixels.
    w : int
        The width of the image in pixels.
    pixels : ndarray
        The 2D array of pixels that make up the mask. Each row is a coordinate
        pair (x, y), such that `coords` has size len(`pixels`) x 2.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import construct_pixel_mask
    >>> h = 3
    >>> w = 3
    >>> pixels = np.array([[0, 0], [1, 1], [2, 1]])
    >>> construct_pixel_mask(h, w, pixels)
    array([[ True, False, False],
           [False,  True,  True],
           [False, False, False]], dtype=bool)

    """

    @_decorate_validation
    def validate_input():
        _numeric('h', 'integer', range_='[2;inf)')
        _numeric('w', 'integer', range_='[2;inf)')
        _numeric('pixels', 'integer', shape=(-1, 2))
        _numeric('pixels[:, 0]', 'integer', shape=(-1,),
                 range_='[0;{})'.format(w), var=pixels[:, 0])
        _numeric('pixels[:, 1]', 'integer', shape=(-1,),
                 range_='[0;{})'.format(h), var=pixels[:, 1])

    validate_input()

    mask = np.zeros((h, w), dtype=np.bool_)
    mask[pixels[:, 1], pixels[:, 0]] = True

    return mask


def sample_lines(coords, speed, sample_rate, time):
    """
    Determine the coordinates of the sampled points in a line scanning.

    Parameters
    ----------
    coords : ndarray
        The `k` floating point coordinates of the line segment endpoints
        arranged into a 2D array where each row is a coordinate pair (x, y),
        such that `coords` has size `k` x 2.
    speed : float
        The probe speed in units of meters/second.
    sample_rate : float
        The sample rate in units of Hertz.
    time : float
        The scan time in units of seconds.

    Returns
    -------
    coords : ndarray
        The coordinates of the samples arranged into a 2D array, such that each
        row is a coordinate pair (x, y).

    """

    dcoords = coords[1:] - coords[:-1]
    cumlen = np.zeros(coords.shape[0])
    cumlen[1:] = np.cumsum(np.sqrt(dcoords[:, 0]**2 + dcoords[:, 1]**2))
    speed = cumlen[-1] / time

    l = np.linspace(0, speed * time, int(sample_rate * time) + 1)
    x = np.interp(l, cumlen, coords[:, 0])
    y = np.interp(l, cumlen, coords[:, 1])

    return np.column_stack((x, y))


def unique_pixels(coords):
    """
    Identify unique pixels from a set of coordinates.

    The floating point `coords` are reduced to a unique set of integer pixels
    by flooring the floating point values.

    Parameters
    ----------
    coords : ndarray
        The `k` floating point coordinates arranged into a 2D array where each
        row is a coordinate pair (x, y), such that `coords` has size `k` x 2.

    Returns
    -------
    unique_pixels : ndarray
        The `l` <= `k` unique (integer) pixels, such that `unique_pixels` is a
        2D array and has size `l` x 2.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import unique_pixels
    >>> coords = np.array([[1.7, 1.0], [1.0, 1.2], [3.3, 4.3]])
    >>> np.int_(unique_pixels(coords))
    array([[1, 1],
           [3, 4]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('coords', ('integer', 'floating'), range_='[0;inf)',
                 shape=(-1, 2))

    validate_input()

    pixels = np.floor(coords).astype(np.int64, order='C')

    pixels_struct = pixels.view(pixels.dtype.descr * 2)
    unique, index = np.unique(pixels_struct, return_index=True)
    unique_pixels = pixels[np.sort(index)]

    return unique_pixels



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing utilities for control of plotting using `matplotlib`.

The module has a number of public attributes which provide settings for
colormap cycles, linestyle cycles, and marker cycles that may be used in
combination with `matplotlib`.

Routine listings
----------------
setup_matplotlib(settings={}, cmap=None)
    Function that set the Magni default `matplotlib` configuration.
colour_collections : dict
    Collections of colours that may be used in e.g., a `matplotlib`
    color_cycle / prop_cycle.
seq_cmaps : list
    Names of `matplotlib.cm` colormaps optimized for sequential data.
div_cmaps : list
    Names of `matplotlib.cm` colormaps optimized for diverging data.
linestyles : list
    A subset of linestyles from `matplotlib.lines`
markers : list
    A subset of markers from `matplotlib.markers`

Examples
--------
Use the default Magni matplotlib settings.

>>> import magni
>>> magni.utils.plotting.setup_matplotlib()

Get the normalised 'Blue' colour brew from the psp colour map:

>>> magni.utils.plotting.colour_collections['psp']['Blue']
((0.1255, 0.502, 0.8745),)

"""

from __future__ import division
import warnings

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from pkg_resources import parse_version as _parse_version

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric

if _parse_version(mpl.__version__) >= _parse_version('1.5.0'):
    import cycler
    _mpl_prop_era = True
else:
    _mpl_prop_era = False


class _ColourCollection(object):
    """
    A container for colour maps.

    A single colour is stored as an RGB 3-tuple of integers in the interval
    [0,255]. A set of related colours is termed a colour brew and is stored as
    a list of colours. A set of related colour brews is termed a colour
    collection and is stored as a dictionary. The dictionary key identifies
    the name of the colour collection whereas the value is the list of colour
    brews.

    The default colour collections named "cb*" are colorblind safe, print
    friendly, and photocopy-able. They have been created using the online
    ColorBrewer 2.0 tool [1]_.

    Parameters
    ----------
    brews : dict
        The dictionary of colour brews from which the colour collection is
        created.

    Notes
    -----
    Each colour brew is a list (or tuple) of length 3 lists (or tuples) of RGB
    values.

    References
    ----------
    .. [1] M. Harrower and C. A. Brewer, "ColorBrewer.org: An Online Tool for
       Selecting Colour Schemes for Maps", *The Cartographic Journal*, vol. 40,
       pp. 27-37, 2003 (See also: http://colorbrewer2.org/)

    """

    def __init__(self, brews):
        @_decorate_validation
        def validate_input():
            _levels('brews', (_generic(None, 'mapping'),
                              _generic(None, 'explicit collection'),
                              _generic(None, 'explicit collection', len_=3),
                              _numeric(None, 'integer', range_='[0;255]')))

        validate_input()

        self._brews = brews

    def __getitem__(self, name):
        """
        Return a single colour brew.

        The returned colour brew is normalised in the sense of matplotlib
        normalised rgb values, i.e., colours are 3-tuples of floats in the
        interval [0, 1].

        Parameters
        ----------
        name : str
            Name of the colour brew to return.

        Returns
        -------
        brew : tuple
            A colour brew list.

        """

        @_decorate_validation
        def validate_input():
            _generic('name', 'string', value_in=tuple(self._brews.keys()))

        validate_input()

        return tuple([tuple([round(val / 255, 4) for val in colour])
                      for colour in self._brews[name]])


colour_collections = {
    'cb4': _ColourCollection({
        'OrRd': ((254, 240, 217), (253, 204, 138), (252, 141, 89),
                 (215, 48, 31)),
        'PuOr': ((230, 97, 1), (253, 184, 99), (178, 171, 210),
                 (94, 60, 153))}),
    'cb3': _ColourCollection({
        'BuGn': ((229, 245, 249), (153, 216, 201), (44, 162, 95)),
        'BuPu': ((224, 236, 244), (158, 188, 218), (136, 86, 167)),
        'GuBu': ((224, 243, 219), (168, 221, 181), (67, 162, 202)),
        'OrRd': ((254, 232, 200), (253, 187, 132), (227, 74, 51)),
        'PuBu': ((236, 231, 242), (166, 189, 219), (43, 140, 190)),
        'PuBuGn': ((236, 226, 240), (166, 189, 219), (28, 144, 153)),
        'PuRd': ((231, 225, 239), (201, 148, 199), (221, 28, 119)),
        'RdPu': ((253, 224, 221), (250, 159, 181), (197, 27, 138)),
        'YlGn': ((247, 252, 185), (173, 221, 142), (49, 163, 84)),
        'YlGnBu': ((237, 248, 177), (127, 205, 187), (44, 127, 184)),
        'YlOrBr': ((255, 247, 188), (254, 196, 79), (217, 95, 14)),
        'YlOrRd': ((255, 237, 160), (254, 178, 76), (240, 59, 32)),
        'Blues': ((222, 235, 247), (158, 202, 225), (49, 130, 189)),
        'Greens': ((229, 245, 224), (161, 217, 155), (49, 163, 84)),
        'Greys': ((240, 240, 240), (189, 189, 189), (99, 99, 99)),
        'Purples': ((239, 237, 245), (188, 189, 220), (117, 107, 177)),
        'Reds': ((254, 224, 210), (252, 146, 114), (222, 45, 38)),
        'PuOr': ((241, 163, 64), (247, 247, 247), (153, 142, 195))}),
    'psp': _ColourCollection({
        'Blue': ((32, 128, 223),),
        'Orange': ((223, 128, 32),),
        'GreenY': ((128, 223, 32),),
        'Purple': ((128, 32, 223),),
        'Red': ((223, 32, 128),),
        'GreenB': ((223, 32, 128),)}),
    'bgg': _ColourCollection({
        'Black': ((0, 0, 0),),
        'Green': ((0, 191, 0),),
        'Grey': ((170, 170, 170),)})}


seq_cmaps = ['YlOrRd', 'YlOrRd_r', 'YlGnBu', 'YlGnBu_r', 'PuBuGn', 'PuBuGn_r',
             'YlOrBr', 'YlOrBr_r', 'BuGn', 'BuGn_r', 'GnBu', 'GnBu_r',
             'PuBu', 'PuBu_r', 'PuRd', 'PuRd_r']
div_cmaps = ['PRGn', 'PRGn_r', 'PiYG', 'PiYG_r', 'RdBu', 'RdBu_r', 'PuOr',
             'PuOr_r', 'RdGy', 'RdGy_r']
linestyles = ['-', '--', '-.', ':']
markers = ['o', '^', 'x', '+', 'd']


def setup_matplotlib(settings={}, cmap=None):
    """
    Adjust the configuration of `matplotlib`.

    Sets the default configuration of `matplotlib` to optimize for producing
    high quality plots of the data produced by the functionality provided in
    the Magni.

    Parameters
    ----------
    settings : dict, optional
       A dictionary of custom matplotlibrc settings. See examples for details
       about the structure of the dictionary.
    cmap : str or tuple, optional
       Colormap to be used by matplotlib (the default is None, which implices
       that the 'coolwarm' colormap is used). If a tuple is supplied it must
       be a ('colormap_name', matplotlib.colors.Colormap()) tuple.

    Raises
    ------
    UserWarning
       If the supplied custom settings are invalid.

    Examples
    --------
    For example, set lines.linewidth=2 and lines.color='r'.

    >>> from magni.utils.plotting import setup_matplotlib
    >>> custom_settings = {'lines': {'linewidth': 2, 'color': 'r'}}
    >>> setup_matplotlib(custom_settings)

    """

    @_decorate_validation
    def validate_input():
        _levels('settings', (_generic(None, 'mapping'),
                             _generic(None, 'mapping')))
        _generic('cmap', ('string', tuple), ignore_none=True)
        if isinstance(cmap, tuple):
            _generic(('cmap', 0), 'string')
            _generic(('cmap', 1), mpl.colors.Colormap)

    validate_input()

    global _settings, _cmap

    for name, setting in settings.items():
        if name in _settings:
            _settings[name].update(setting)
        else:
            _settings[name] = setting

    for name, setting in _settings.items():
        try:
            mpl.rc(name, **setting)
        except (AttributeError, KeyError):
            warnings.warn('Setting {!r} ignored.'.format(name), UserWarning)

    if cmap is not None:
        if isinstance(cmap, tuple):
            mpl.cm.register_cmap(name=cmap[0], cmap=cmap[1])
            cmap = cmap[0]
        plt.set_cmap(cmap)
    elif _cmap is not None:
        plt.set_cmap(_cmap)

    _settings = {}
    _cmap = None

if _mpl_prop_era:
    # Matplotlib >= 1.5.0
    _style_cycle = cycler.cycler('linestyle', linestyles)
    _color_cycle = cycler.cycler('color', colour_collections['cb4']['PuOr'])
    _prop_settings = {'axes': {'prop_cycle': _style_cycle * _color_cycle}}

else:
    _prop_settings = {
        'axes': {'color_cycle': colour_collections['cb4']['PuOr']}}

_settings = dict({'text': {'usetex': False},
                  'font': {'size': 12},
                  'mathtext': {'fontset': 'cm'},
                  'pdf': {'fonttype': 42},
                  'ps': {'fonttype': 42},
                  'legend': {'fontsize': 11},
                  'lines': {'linewidth': 2},
                  'figure': {
                      'figsize': (8.0, float(8.0 / ((1 + np.sqrt(5)) / 2))),
                      'dpi': 600},
                  'image': {'interpolation': 'none'}}, **_prop_settings)
_cmap = 'coolwarm'



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing functionality for image manipulation.

Routine listings
----------------
dictionaries
    Module providing fast linear operations wrapped in matrix emulators.
domains
    Module providing a multi domain image class.
evaluation
    Module providing functions for evaluation of image reconstruction quality.
measurements
    Module providing functions for constructing scan patterns for measurements.
preprocessing
    Module providing functionality to remove tilt in images.
visualisation
    Module providing functionality for visualising images.
mat2vec(x)
    Function to reshape a matrix into vector by stacking columns.
vec2mat(x, mn_tuple)
    Function to reshape a vector into a matrix.
double_mirror(img, fftstyle=True)
    Function to mirror image in both the vertical and horisontal axes.
get_inscribed_masks(img, as_vec=False)
    Function to get inscribed masks covering the image.

Notes
-----
See `_util` for documentation of `mat2vec`, `vec2mat`, `double_mirror`, and
`get_inscribed_masks`.

"""

# the mat2vec and vec2mat functions need to be imported first to avoid
# recursive imports
from magni.imaging._util import mat2vec
from magni.imaging._util import vec2mat
from magni.imaging._util import double_mirror
from magni.imaging._util import get_inscribed_masks

from magni.imaging import dictionaries
from magni.imaging import domains
from magni.imaging import evaluation
from magni.imaging import measurements
from magni.imaging import preprocessing
from magni.imaging import visualisation



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing matrix emulators.

The matrix emulators of this module are wrappers of fast linear operations
giving the fast linear operations the same basic interface as a numpy ndarray.
Thereby allowing fast linear operations and numpy ndarrays to be used
interchangably in other parts of the package.

Routine listings
----------------
Matrix(magni.utils.validation.types.MatrixBase)
    Wrap fast linear operations in a matrix emulator.
MatrixCollection(magni.utils.validation.types.MatrixBase)
    Wrap multiple matrix emulators in a single matrix emulator.
Separable2DTransform(Matrix)
    Wrap a linear 2D separable transform in a matrix emulator.
SumApproximationMatrix(object)
    Wrap a sum approximation in a matrix emulator.
norm(A, ord=None)
    Compute a norm of a matrix.

See Also
--------
magni.imaging._fastops : Fast linear operations.
magni.imaging._mtx1D : 1D transforms matrices for use 2D separable transforms.

"""

from __future__ import division
import copy

import numpy as np

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation.types import MatrixBase as _MatrixBase
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric
from magni.utils.validation import validate_once as _validate_once


class Matrix(_MatrixBase):
    """
    Wrap fast linear operations in a matrix emulator.

    `Matrix` defines a few attributes and internal methods which ensures that
    instances have the same basic interface as a numpy ndarray instance without
    explicitly forming the array. This basic interface allows instances to be
    multiplied with vectors, to be transposed, to be complex conjuagted, and to
    assume a shape. Also, instances have an attribute which explicitly forms
    the matrix as an ndarray.

    Parameters
    ----------
    func : function
        The fast linear operation applied to the vector when multiplying the
        matrix with a vector.
    conj_trans : function
        The fast linear operation applied to the vector when multiplying the
        complex conjuated transposed matrix with a vector.
    args : list or tuple
        The arguments which should be passed to `func` and `conj_trans` in
        addition to the vector.
    shape : list or tuple
        The shape of the emulated matrix.
    is_complex : bool
        The indicator of whether or not the emulated matrix is defined for the
        complex numbers (the default is False, which implies that the emulated
        matrix is defined for the real numbers only).
    is_valid : bool
        The indicator of whether or not the fast linear operation corresponds
        to a valid matrix (se discussion below) (the default is True, which
        implies that the matrix is considered valid).

    See Also
    --------
    magni.utils.validation.types.MatrixBase : Superclass of the present class.

    Notes
    -----
    The `is_valid` indicator is an implementation detail used to control
    possibly missing operations in the fast linear operation. For instance,
    consider the Discrete Fourier Transform (DFT) and its inverse transform.
    The forward DFT transform is a matrix-vector product involving the
    corresponding DFT matrix. The inverse transform is a matrix-vetor product
    involving the complex conjugate transpose DFT matrix. That is, it involves
    complex conjugation and transposition, both of which are (individually)
    valid transformations of the DFT matrix. However, when implementing the DFT
    (and its inverse) using a Fast Fourier Transform (FFT), only the combined
    (complex conjugate, transpose) operation is available. Thus, when using an
    FFT based `magni.util.matrices.Matrix`, one can get, e.g., a `Matrix.T`
    object corresponding to its transpose. However, the operation of computing
    a matrix-vector product involving the tranpose DFT matrix is not available
    and the `Matrix.T` is, consequently, considered an invalid matrix. Only the
    combined `Matrix.T.conj()` or `Matrix.conj().T` is considered a valid
    matrix.

    .. warning:: The tranpose operation changed in `magni` 1.5.0.

        For complex matrices, the the `.T` tranpose operation now yields an
        invalid matrix as described above. Prior to `magni` 1.5.0, the `.T`
        would yield the "inverse" which would usually be the complex conjugated
        transpose for complex matrices.


    Examples
    --------
    For example, the negative identity matrix could be emulated as

    >>> import numpy as np, magni
    >>> from magni.utils.matrices import Matrix
    >>> func = lambda vec: -vec
    >>> matrix = Matrix(func, func, (), (3, 3))

    The example matrix will have the desired shape:

    >>> matrix.shape
    (3, 3)

    The example matrix will behave just like an explicit matrix:

    >>> vec = np.float64([1, 2, 3]).reshape(3, 1)
    >>> np.set_printoptions(suppress=True)
    >>> matrix.dot(vec)
    array([[-1.],
           [-2.],
           [-3.]])

    If, at some point, an explicit representation of the matrix is required,
    this can easily be obtained:

    >>> matrix.A
    array([[-1., -0., -0.],
           [-0., -1., -0.],
           [-0., -0., -1.]])

    Likewise, the transpose of the matrix can be obtained:

    >>> matrix.T.A
    array([[-1., -0., -0.],
           [-0., -1., -0.],
           [-0., -0., -1.]])

    """

    def __array__(self):
        """Return ndarray representation of the matrix."""
        return self.A

    def __init__(self, func, conj_trans, args, shape, is_complex=False,
                 is_valid=True):
        _MatrixBase.__init__(self, np.complex_ if is_complex else np.float_,
                             ((-np.inf, np.inf), (-np.inf, np.inf)), shape)

        @_decorate_validation
        def validate_input():
            _generic('func', 'function')
            _generic('conj_trans', 'function')
            _generic('args', 'explicit collection')
            _levels('shape', (_generic(None, 'explicit collection', len_=2),
                              _numeric(None, 'integer')))
            _numeric('is_complex', 'boolean')
            _numeric('is_valid', 'boolean')

        validate_input()

        self._func = func
        self._conj_trans = conj_trans
        self._args = tuple(args)
        self._is_complex = is_complex
        self._is_valid = is_valid

    @property
    def matrix_state(self):
        """
        Return a copy of the internal matrix state.

        The internal matrix state consists of:

        * func: The forward fast linear operator.
        * conj_trans: The backward fast linear operator.
        * args: The tuple of extra arguments passed to `func` and `conj_trans`.
        * is_complex: The indicator of whether or not the matrix is complex.
        * is_valid: The indicator of whether or not the matrix is valid.

        Returns
        -------
        matrix_state : dict
            A copy of the internal matrix state

        """

        matrix_state = {'func': self._func,
                        'conj_trans': self._conj_trans,
                        'args': copy.deepcopy(self._args),
                        'is_complex': self._is_complex,
                        'is_valid': self._is_valid}

        return matrix_state

    @property
    def A(self):
        """
        Explicitly form the matrix.

        The fast linear operations implicitly define a matrix which is usually
        not explicitly formed. However, some functionality might require a more
        advanced matrix interface than that provided by this class.

        Returns
        -------
        matrix : numpy.ndarray
            The explicit matrix.

        Notes
        -----
        The explicit matrix is formed by multiplying the matrix with the
        columns of an identity matrix and stacking the resulting vectors as
        columns in a matrix.

        """

        if not self._is_valid:
            raise ValueError('This matrix has not been implemented. Transpose '
                             'or conjugate it for an implemented matrix.')

        output = np.zeros(
            self.shape, dtype=np.complex_ if self._is_complex else np.float_)
        vec = np.zeros((self.shape[1], 1))

        for i in range(self.shape[1]):
            vec[i] = 1
            output[:, i] = self.dot(vec)[:, 0]
            vec[i] = 0

        return output

    @property
    def T(self):
        """
        Get the transpose of the matrix.

        Returns
        -------
        matrix : Matrix
            The transpose of the matrix.

        Notes
        -----
        The fast linear operation and the fast linear transposed operation of
        the resulting matrix are the same as those of the current matrix except
        swapped. The shape is modified accordingly. This returns an `invalid`
        matrix if the entries are complex numbers as only the complex conjugate
        transpose is considered valid.

        """

        if self._is_complex:
            is_valid = not self._is_valid
        else:
            is_valid = True

        return Matrix(self._conj_trans, self._func, self._args,
                      self._shape[::-1], self._is_complex, is_valid)

    def conj(self):
        """
        Get the complex conjugate of the matrix.

        Returns
        -------
        matrix : Matrix
            The complex conjugate of the matrix.

        Notes
        -----
        The fast linear operation and the fast linear transposed operation of
        the resulting matrix are the same as those of the current matrix.
        This returns an `invalid` matrix if the entries are complex numbers as
        only the complex conjugate transpose is considered valid.

        """

        if self._is_complex:
            is_valid = not self._is_valid
        else:
            is_valid = True

        return Matrix(self._func, self._conj_trans, self._args, self._shape,
                      self._is_complex, is_valid)

    @_validate_once
    def dot(self, vec):
        """
        Multiply the matrix with a vector.

        Parameters
        ----------
        vec : numpy.ndarray
            The vector which the matrix is multiplied with.

        Returns
        -------
        vec : numpy.ndarray
            The result of the multiplication.

        Notes
        -----
        This method honors `magni.utils.validation.enable_validate_once`.

        """

        @_decorate_validation
        def validate_input():
            _numeric('vec', ('integer', 'floating', 'complex'),
                     shape=(self.shape[1], 1))

        validate_input()

        if not self._is_valid:
            raise ValueError('This matrix has not been implemented. Transpose '
                             'or conjugate it for an implemented matrix.')

        return self._func(vec, *self._args)


class MatrixCollection(_MatrixBase):
    """
    Wrap multiple matrix emulators in a single matrix emulator.

    `MatrixCollection` defines a few attributes and internal methods which
    ensures that instances have the same basic interface as a numpy ndarray
    instance without explicitly forming the ndarray. This basic interface
    allows instances to be multiplied with vectors, to be transposed, to be
    complex conjugated, and to assume a shape. Also, instances have an
    attribute which explicitly forms the matrix.

    Parameters
    ----------
    matrices : list or tuple
        The collection of matrices (e.g. ndarrays or `Matrix` instances).

    See Also
    --------
    magni.utils.validation.types.MatrixBase : Superclass of the present class.
    Matrix : Matrix emulator.

    Examples
    --------
    For example, two matrix emulators can be combined into one. That is, the
    matrix:

    >>> import numpy as np, magni
    >>> func = lambda vec: -vec
    >>> negate = magni.utils.matrices.Matrix(func, func, (), (3, 3))
    >>> np.set_printoptions(suppress=True)
    >>> negate.A
    array([[-1., -0., -0.],
           [-0., -1., -0.],
           [-0., -0., -1.]])

    And the matrix:

    >>> func = lambda vec: vec[::-1]
    >>> reverse = magni.utils.matrices.Matrix(func, func, (), (3, 3))
    >>> reverse.A
    array([[ 0.,  0.,  1.],
           [ 0.,  1.,  0.],
           [ 1.,  0.,  0.]])

    Can be combined into one matrix emulator using the present class:

    >>> from magni.utils.matrices import MatrixCollection
    >>> matrix = MatrixCollection((negate, reverse))

    The example matrix will have the desired shape:

    >>> matrix.shape
    (3, 3)

    The example matrix will behave just like an explicit matrix:

    >>> vec = np.float64([1, 2, 3]).reshape(3, 1)
    >>> matrix.dot(vec)
    array([[-3.],
           [-2.],
           [-1.]])

    If, at some point, an explicit representation of the matrix is required,
    this can easily be obtained:

    >>> matrix.A
    array([[-0., -0., -1.],
           [-0., -1., -0.],
           [-1., -0., -0.]])

    Likewise, the transpose of the matrix can be obtained:

    >>> matrix.T.A
    array([[-0., -0., -1.],
           [-0., -1., -0.],
           [-1., -0., -0.]])

    """

    def __array__(self):
        """Return ndarray representation of the matrix."""
        return self.A

    def __init__(self, matrices):
        _MatrixBase.__init__(self, np.complex_,
                             ((-np.inf, np.inf), (-np.inf, np.inf)), None)

        @_decorate_validation
        def validate_input():
            _levels('matrices', (_generic(None, 'explicit collection'),
                                 _numeric(
                                     None,
                                     ('integer', 'floating', 'complex'),
                                     shape=(-1, -1))))

            for i in range(len(matrices) - 1):
                if matrices[i].shape[1] != matrices[i + 1].shape[0]:
                    msg = ('The value of >>matrices[{}].shape[1]<<, {!r}, '
                           'must be equal to the value of '
                           '>>matrices[{}].shape[0]<<, {!r}.')
                    raise ValueError(msg.format(i, matrices[i].shape[1], i + 1,
                                                matrices[i + 1].shape[0]))

        validate_input()

        self._matrices = tuple(matrices)

    @property
    def matrix_state(self):
        """
        Return a copy of the internal matrix state.

        The internal matrix state consists of:

        * matrices: The tuple of matrices in the matrix collection

        Returns
        -------
        matrix_state : dict
            A copy of the internal matrix state

        """

        matrix_state = {'matrices': copy.deepcopy(self._matrices)}

        return matrix_state

    @property
    def A(self):
        """
        Explicitly form the matrix.

        The collection of matrices implicitly defines a matrix which is usually
        not explicitly formed. However, some functionality might require a more
        advanced matrix interface than that provided by this class.

        Returns
        -------
        matrix : numpy.ndarray
            The explicit matrix.

        Notes
        -----
        The explicit matrix is formed by multiplying the matrix with the
        columns of an identity matrix and stacking the resulting vectors as
        columns in a matrix.

        """

        is_complex = any([getattr(matrix, '_is_complex', True)  # Default to
                          for matrix in self._matrices])       # complex matrix
        output = np.zeros(
            self.shape, dtype=np.complex_ if is_complex else np.float_)
        vec = np.zeros((self.shape[1], 1))

        for i in range(self.shape[1]):
            vec[i] = 1
            output[:, i] = self.dot(vec)[:, 0]
            vec[i] = 0

        return output

    @property
    def shape(self):
        """
        Get the shape of the matrix.

        The shape of the product of a number of matrices is the number of rows
        of the first matrix times the number of columns of the last matrix.

        Returns
        -------
        shape : tuple
            The shape of the matrix.

        """

        return (self._matrices[0].shape[0], self._matrices[-1].shape[1])

    @property
    def T(self):
        """
        Get the transpose of the matrix.

        The transpose of the product of the number of matrices is the product
        of the transpose of the matrices in reverse order.

        Returns
        -------
        matrix : MatrixCollection
            The transpose of the matrix.

        """

        return MatrixCollection([matrix.T for matrix in self._matrices[::-1]])

    def conj(self):
        """
        Get the complex conjugate of the matrix.

        The complex conjugate of the product of the number of matrices is the
        product of the complex conjugates of the matrices.

        Returns
        -------
        matrix : MatrixCollection
            The complex conjugate of the matrix.

        """

        return MatrixCollection([matrix.conj() for matrix in self._matrices])

    @_validate_once
    def dot(self, vec):
        """
        Multiply the matrix with a vector.

        Parameters
        ----------
        vec : numpy.ndarray
            The vector which the matrix is multiplied with.

        Returns
        -------
        vec : numpy.ndarray
            The result of the multiplication.

        Notes
        -----
        This method honors `magni.utils.validation.enable_validate_once`.

        """

        @_decorate_validation
        def validate_input():
            _numeric('vec', ('integer', 'floating', 'complex'),
                     shape=(self.shape[1], 1))

        validate_input()

        for matrix in self._matrices[::-1]:
            vec = matrix.dot(vec)

        return vec


class Separable2DTransform(Matrix):
    """
    Wrap a linear 2D separable transform in a matrix emulator.

    A linear 2D separable transform is defined by the two matrices `mtx_l` and
    `mtx_r` in the sense that the Kronecker product kron(`mtx_l`, `mtx_r`)
    yields the full matrix of the linear 2D separable transform when the linear
    transform is implemented as a matrix-vector product. See e.g. [1]_ for the
    details.

    `Separable2DTransform` defines a few attributes and internal methods which
    ensures that instances have the same basic interface as a numpy matrix
    instance without explicitly forming the matrix. This basic interface allows
    instances to be multiplied with vectors, to be transposed, and to assume a
    shape. Also, instances have an attribute which explicitly forms the matrix.

    Parameters
    ----------
    mtx_l : ndarray
        The "left" matrix in the defining Kronecker product.
    mtx_r : ndarray
        The "right" matrix in the defining Kronecker product.

    See Also
    --------
    magni.utils.matrices.Matrix : Superclass of the present class.

    References
    ----------
    .. [1] A.N. Akansu, R.A. Haddad, and P.R. Haddad, *Multiresolution Signal
       Decomposition: Transforms, Subbands, and Wavelets*, Academic Press,
       2000.

    Examples
    --------
    For example, the transform based on a 2-by-3 matrix and a 5-by-4 matrix

    >>> import numpy as np
    >>> from magni.imaging import vec2mat, mat2vec
    >>> from magni.utils.matrices import Separable2DTransform
    >>> mtx_l = np.arange(6).reshape(2, 3)
    >>> mtx_r = np.arange(40)[::2].reshape(5, 4)
    >>> sep_matrix = Separable2DTransform(mtx_l, mtx_r)

    The full transform matrix shape is:

    >>> print(tuple(int(s) for s in sep_matrix.shape))
    (10, 12)

    The matrix behaves like an explicit matrix

    >>> vec = np.arange(36)[::3].reshape(12, 1)
    >>> sep_matrix.dot(vec)
    array([[  936],
           [ 3528],
           [ 2712],
           [10056],
           [ 4488],
           [16584],
           [ 6264],
           [23112],
           [ 8040],
           [29640]])

    which, due to the separability of the transform, may also be computed as

    >>> vec_as_mat = vec2mat(vec, (3, 4))
    >>> mat2vec(mtx_l.dot(vec_as_mat.dot(mtx_r.T)))
    array([[  936],
           [ 3528],
           [ 2712],
           [10056],
           [ 4488],
           [16584],
           [ 6264],
           [23112],
           [ 8040],
           [29640]])

    The explicit matrix is given by the kronecker product of the two matrices

    >>> np.all(sep_matrix.A == np.kron(mtx_l, mtx_r))
    True

    The transpose of the matrix is also easily obtainable

    >>> sep_matrix_transpose = sep_matrix.T
    >>> sep_matrix_transpose.A
    array([[  0,   0,   0,   0,   0,   0,  24,  48,  72,  96],
           [  0,   0,   0,   0,   0,   6,  30,  54,  78, 102],
           [  0,   0,   0,   0,   0,  12,  36,  60,  84, 108],
           [  0,   0,   0,   0,   0,  18,  42,  66,  90, 114],
           [  0,   8,  16,  24,  32,   0,  32,  64,  96, 128],
           [  2,  10,  18,  26,  34,   8,  40,  72, 104, 136],
           [  4,  12,  20,  28,  36,  16,  48,  80, 112, 144],
           [  6,  14,  22,  30,  38,  24,  56,  88, 120, 152],
           [  0,  16,  32,  48,  64,   0,  40,  80, 120, 160],
           [  4,  20,  36,  52,  68,  10,  50,  90, 130, 170],
           [  8,  24,  40,  56,  72,  20,  60, 100, 140, 180],
           [ 12,  28,  44,  60,  76,  30,  70, 110, 150, 190]])

    """

    def __init__(self, mtx_l, mtx_r):
        _MatrixBase.__init__(self, np.complex_,
                             ((-np.inf, np.inf), (-np.inf, np.inf)),
                             (mtx_l.shape[0] * mtx_r.shape[0],
                              mtx_l.shape[1] * mtx_r.shape[1]))

        @_decorate_validation
        def validate_input():
            _numeric('mtx_l', ('integer', 'floating', 'complex'),
                     shape=(-1, -1))
            _numeric('mtx_r', ('integer', 'floating', 'complex'),
                     shape=(-1, -1))

        validate_input()

        self._mtx_l = mtx_l
        self._mtx_r = mtx_r
        self._is_complex = np.iscomplexobj(mtx_l) or np.iscomplexobj(mtx_r)

    @property
    def matrix_state(self):
        """
        Return a copy of the internal matrix state.

        The internal matrix state consists of:

        * mtx_l: The "left" matrix in the defining Kronecker product.
        * mtx_r: The "right" matrix in the defining Kronecker product.

        Returns
        -------
        matrix_state : dict
            A copy of the internal matrix state

        """

        matrix_state = {'mtx_l': self._mtx_l.copy(),
                        'mtx_r': self._mtx_r.copy()}

        return matrix_state

    @property
    def A(self):
        """
        Explicitly form the matrix.

        For a linear separable 2D transform, the full explicit transform matrix
        is given by the Kronecker product of the two matrices that define the
        separable transform.

        Returns
        -------
        matrix : ndarray
            The explicit matrix.

        """

        return np.kron(self._mtx_l, self._mtx_r)

    @property
    def T(self):
        """
        Get the transpose of the matrix.

        The transpose of a Kronekcer product of two matrices is the Kronecker
        product of the transposes of the two matrices.

        Returns
        -------
        matrix : Separable2DTransform
            The transpose of the matrix.

        """

        return Separable2DTransform(self._mtx_l.T, self._mtx_r.T)

    def conj(self):
        """
        Get the complex conjugate of the matrix.

        The complex conjugate of a Kronekcer product of two matrices is the
        Kronecker product of the complex conjugates of the two matrices.

        Returns
        -------
        matrix : Separable2DTransform
            The complex conjugate of the matrix.

        """

        return Separable2DTransform(self._mtx_l.conj(), self._mtx_r.conj())

    @_validate_once
    def dot(self, vec):
        """
        Multiply the matrix with a vector.

        The matrix-vector product is efficiently computed by
        mtx_l.dot(V.dot(mtx_r.T)) where V is the matrix from which `vec` was
        created by stacking its columns.

        Parameters
        ----------
        vec : ndarray
            The vector corresponding to the stacked columns of a matrix which
            this matrix is multiplied with.

        Returns
        -------
        vec : ndarray
            The result of the matrix-vector multiplication.

        Notes
        -----
        This method honors `magni.utils.validation.enable_allow_validate_once`.

        """

        @_decorate_validation
        def validate_input():
            _numeric('vec', ('integer', 'floating', 'complex'),
                     shape=(self.shape[1], 1))

        validate_input()

        vec_as_mat = vec.reshape(self._mtx_r.shape[1], self._mtx_l.shape[1]).T
        mat_mult_result = self._mtx_l.dot(vec_as_mat.dot(self._mtx_r.T))
        mat_mult_result_as_vec = mat_mult_result.T.reshape(-1, 1)

        return mat_mult_result_as_vec


class SumApproximationMatrix(object):
    """
    Wrap a sum approximation in a matrix emulator.

    This simply emulates computing a scaled sum of the entries of a vector as a
    matrix vector product.

    Parameters
    ----------
    scaling : int or float
        The scaling applied to the sum approximation.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.utils.matrices import SumApproximationMatrix
    >>> np.random.seed(seed=6021)
    >>> n = 10
    >>> vec = np.random.randn(n, 1)
    >>> np_printoptions = np.get_printoptions()
    >>> np.set_printoptions(precision=5)
    >>> np.array([vec.mean()])
    array([ 0.43294])
    >>> matrix = SumApproximationMatrix(1.0/n)
    >>> np.array([matrix.dot(vec)])
    array([ 0.43294])
    >>> np.set_printoptions(**np_printoptions)

    """

    def __init__(self, scaling):
        @_decorate_validation
        def validate_input():
            _numeric('scaling', ('integer', 'floating'))

        validate_input()

        self.scaling = scaling

    @_validate_once
    def dot(self, vec):
        """
        Form the matrix-vector product sum approximation.

        Parameters
        ----------
        vec : ndarray
            The vector in the matrix-vector product.

        Returns
        -------
        scaled sum : float
            The scaled sum of the input vector.

        Notes
        -----
        This method honors `magni.utils.validation.enable_allow_validate_once`.

        """

        @_decorate_validation
        def validate_input():
            _numeric('vec', ('integer', 'floating', 'complex'), shape=(-1, 1))

        validate_input()

        return self.scaling * np.sum(vec)


def norm(A, ord=None):
    """
    Compute a norm of a matrix.

    Efficient norm computations for the `magni.utils.matrices` matrices.

    Parameters
    ----------
    A : magni.util.matrices.Matrix or magni.utils.matrices.MatrixCollection
        The matrix which norm is to be computed.
    ord : str
        The order of the norm to compute.

    Returns
    -------
    norm : float
        The computed matrix norm.

    See Also
    --------
    numpy.linalg.norm : Numpy's function for computing norms of ndarrays

    Notes
    -----
    This function is a simple wrapper around `numpy.linalg.norm`. It exploits
    some properties of the matrix classes in `magni.utils.matrices` to optimize
    the norm computation in terms of speed and/or memory usage.

    .. warning::

        This function builds the ndarray corresponding to the `A` matrix and
        passes that on to `numpy.linalg.norm` if no smarter way of computing
        the norm has been implemented. Beware of possible "out of memory"
        issues.

    Examples
    --------
    Compute the Frobenius norm of a Seperable2DTransform

    >>> import numpy as np
    >>> from magni.utils.matrices import Separable2DTransform, norm
    >>> mtx_l = np.arange(6).reshape(2, 3)
    >>> mtx_r = np.arange(40)[::2].reshape(5, 4)
    >>> sep_matrix = Separable2DTransform(mtx_l, mtx_r)
    >>> round(float(norm(sep_matrix, 'fro')), 2)
    737.16

    for comparison, the Frobenius computed using np.linalg.norm is

    >>> round(float(np.linalg.norm(sep_matrix.A)), 2)
    737.16

    """

    @_decorate_validation
    def validate_input():
        _numeric('A', ('integer', 'floating', 'complex'), shape=(-1, -1))
        _generic('ord', 'string', ignore_none=True)

    validate_input()

    norm = None

    # Special computations
    if ord is None or ord == 'fro':
        if isinstance(A, Separable2DTransform):
            A_s = A.matrix_state
            norm = np.linalg.norm(A_s['mtx_l']) * np.linalg.norm(A_s['mtx_r'])

    # Fallback numpy solution
    if norm is None:
        norm = np.linalg.norm(A)

    return norm



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the `validate_numeric` function.

Routine listings
----------------
validate_numeric(name, type, range_='[-inf;inf]', shape=(), precision=None,
    ignore_none=False, var=None)
    Validate numeric objects.

"""

from __future__ import division
from numbers import Number as _Number

import numpy as np

from magni.utils.validation.types import MatrixBase as _MatrixBase
from magni.utils.validation._util import get_var as _get_var
from magni.utils.validation._util import report as _report


try:
    long = long
except NameError:
    long = int

_types = {
    'boolean': {
        None: (bool, np.bool_),
        8: np.bool8 if 'bool8' in dir(np) else np.bool_},
    'integer': {
        None: (int, np.signedinteger, long),
        8: getattr(np, 'int8', np.int_),
        16: getattr(np, 'int16', np.int_),
        32: getattr(np, 'int32', np.int_),
        64: getattr(np, 'int64', np.int_)},
    'floating': {
        None: (float, np.floating),
        16: getattr(np, 'float16', np.float_),
        32: getattr(np, 'float32', np.float_),
        64: getattr(np, 'float64', np.float_),
        128: getattr(np, 'float128', np.float_)},
    'complex': {
        None: (complex, np.complexfloating),
        32: getattr(np, 'complex64', np.complex_),
        64: getattr(np, 'complex128', np.complex_),
        128: getattr(np, 'complex256', np.complex_)}}

_precisions = set(precision
                  for type_ in _types.values()
                  for precision in type_.keys())


def validate_numeric(name, type_, range_='[-inf;inf]', shape=(),
                     precision=None, ignore_none=False, var=None):
    """
    Validate numeric objects.

    The present function is meant to valdiate the type or class of an object.
    Furthermore, if the object may only take on a connected range of values,
    the object can be validated against this range. Also, the shape of the
    object can be validated. Finally, the precision used to represent the
    object can be validated.

    If the present function is called with `name` set to None, an iterable with
    the value 'numeric' followed by the remaining arguments passed in the call
    is returned. This is useful in combination with the validation function
    `magni.utils.validation.validate_levels`.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    type_ : None
        One or more references to groups of types.
    range_ : None
        The range of accepted values. (the default is '[-inf;inf]', which
        implies that all values are accepted)
    shape : list or tuple
        The accepted shape. (the default is (), which implies that only scalar
        values are accepted)
    precision : None
        One or more precisions.
    ignore_none : bool
        A flag indicating if the variable is allowed to be none. (the default
        is False)
    var : None
        The value of the variable to be validated.

    See Also
    --------
    magni.utils.validation.validate_generic : Validate non-numeric objects.
    magni.utils.validation.validate_levels : Validate contained objects.

    Notes
    -----
    `name` must refer to a variable in the parent scope of the function or
    method decorated by `magni.utils.validation.decorate_validation` which is
    closest to the top of the call stack. If `name` is a string then there must
    be a variable of that name in that scope. If `name` is a set-like object
    then there must be a variable having the first value in that set-like
    object as name. The remaining values are used as keys/indices on the
    variable to obtain the variable to be validated. For example, the `name`
    ('name', 0, 'key') refers to the variable "name[0]['key']".

    `type_` is either a single value treated as a list with one value or a
    set-like object containing at least one value. Each value refers to a
    number of data types depending if the string value is 'boolean', 'integer',
    'floating', or 'complex'.

    - 'boolean' tests if the variable is a bool or has the data type
      `numpy.bool8`.
    - 'integer' tests if the variable is an int or has the data type
      `numpy.int8`, `numpy.int16`, `numpy.int32`, or `numpy.int64`.
    - 'floating' tests if the variable is a float or has the data type
      `numpy.float16`, `numpy.float32`, `numpy.float64`, or `numpy.float128`.
    - 'complex' tests if the variable is a complex or has the data type
      `numpy.complex32`, `numpy.complex64`, or `numpy.complex128`.

    Because `bool` is a subclass of `int`, a `bool` will pass validation as an
    'integer'. This, however, is not the case for `numpy.bool8`.

    `range_` is either a list with two strings or a single string. In the
    latter case, the default value of the argument is used as the second
    string. The first value represents the accepted range of real values
    whereas the second value represents the accepted range of imaginary values.
    Each string consists of the following parts:

    - One of the following delimiters: '[', '(', ']'.
    - A numeric value (or '-inf').
    - A semi-colon.
    - A numeric value (or 'inf').
    - One of the following delimiters: ']', ')', '['.

    `shape` is either None meaning that any shape is accepted or a list of
    integers. In the latter case, the integer -1 may be used to indicate that
    the given axis may have any length.

    `precision` is either an integer treated as a list with one value or a
    list or tuple containing at least one integer. Each value refers to an
    accepted number of bits used to store each value of the variable.

    `var` can be used to pass the value of the variable to be validated. This
    is useful either when the variable cannot be looked up by `name` (for
    example, if the variable is a property of the argument of a function) or to
    remove the overhead of looking up the value.

    Examples
    --------
    Every public function and method of the present package (with the exception
    of the functions of this subpackage itself) validates every argument and
    keyword argument using the functionality of this subpackage. Thus, for
    examples of how to use the present function, browse through the code.

    """

    if name is None:
        return ('numeric', type_, range_, shape, precision, ignore_none)

    if var is None:
        var = _get_var(name)

    if var is None:
        if not isinstance(ignore_none, bool):
            _report(TypeError, 'must be {!r}.', bool, var_name='ignore_none',
                    var_value=ignore_none, expr='type({})',
                    prepend='Invalid validation call: ')

        if ignore_none:
            return
        else:
            _report(ValueError, 'must not be {!r}.', None, var_name=name)

    if not isinstance(var, (_Number, np.generic, np.ndarray, _MatrixBase)):
        _report(TypeError, '>>{}<<, {!r}, must be numeric.', (name, var),
                prepend='The value(s) of ')

    dtype, bounds, dshape = _examine_var(name, var)

    if isinstance(type_, str) or not hasattr(type_, '__iter__'):
        type_ = (type_,)

    _check_type(name, dtype, type_)
    _check_range(name, bounds, range_)
    _check_shape(name, dshape, shape)
    _check_precision(name, dtype, type_, precision)


def _check_precision(name, dtype, types_, precision):
    """
    Check the precision of a variable.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    dtype : type
        The data type of the variable to be validated.
    types_ : set-like
        The list of accepted types.
    precision : None
        The accepted precision(s).

    """

    if isinstance(precision, (list, tuple)):
        precisions = precision
    else:
        precisions = (precision,)

    for precision in precisions:
        if precision not in _precisions:
            _report(ValueError, 'must be in {!r}.', _precisions,
                    var_name='precision', var_value=precisions,
                    prepend='Invalid validation call: ')

    if None not in precisions:
        dtypes = [_types[type_][precision_]
                  for type_ in types_
                  for precision_ in precisions
                  if precision_ in _types[type_]]

        if not isinstance(dtype(), tuple(dtypes)):
            _report(TypeError, '>>{}.dtype<<, {!r}, must be in {!r}.',
                    (name, dtype, dtypes), prepend='The value(s) of ')


def _check_range(name, bounds, range_):
    """
    Check the range of a variable.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    bounds : list or tuple
        The bounds of the variable to be validated.
    range_ : None
        The accepted range(s).

    """

    if not isinstance(range_, (list, tuple)):
        range_ = (range_, '[-inf;inf]')

    try:
        tests = [req.split(';') for req in range_]

        for i in (0, 1):
            tests[i][0] = [bounds[i][0], tests[i][0][0], tests[i][0][1:]]
            tests[i][0][1] = {']': '>', '(': '>', '[': '>='}[tests[i][0][1]]
            tests[i][1] = [bounds[i][1], tests[i][1][-1], tests[i][1][:-1]]
            tests[i][1][1] = {'[': '<', ')': '<', ']': '<='}[tests[i][1][1]]

        for expr, tests_ in zip(('real({})', 'imag({})'), tests):
            for func, test in zip(('min', 'max'), tests_):
                bound, op, value = test
                test = '{} {} {}'.format(bound, op, value)

                if not eval(test.replace('inf', 'np.inf')):
                    _report(ValueError, '>>{}({})<<, {!r}, must be {} {!r}.',
                            (func, expr.format(name), bound, op, eval(value)),
                            prepend='The value(s) of ')
    except (AttributeError, KeyError, IndexError, SyntaxError):
        _report(ValueError, 'must be a valid range.', var_name='range_',
                var_value=range_, prepend='Invalid validation call: ')


def _check_shape(name, dshape, shape):
    """
    Check the shape of a variable.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    dshape : list or tuple
        The shape of the variable to be validated.
    shape : list or tuple
        The accepted shape.

    """

    if shape is not None:
        if not isinstance(shape, (list, tuple)):
            _report(TypeError, 'must be in {!r}.', (list, tuple),
                    var_name='shape', var_value=shape, expr='type({})',
                    prepend='Invalid validation call: ')

        if len(dshape) != len(shape):
            _report(ValueError, '>>len({}.shape)<<, {!r}, must be {!r}.',
                    (name, len(dshape), len(shape)),
                    prepend='The value(s) of ')

        for i, value in enumerate(shape):
            if not isinstance(value, (int, long)):
                _report(TypeError, 'must be {!r}.', (int, long),
                        var_name=('shape', i), var_value=value,
                        expr='type({})', prepend='Invalid validation call: ')

            if value > -1 and dshape[i] != value:
                _report(ValueError, '>>{}.shape[{}]<<, {!r}, must be {!r}.',
                        (name, i, dshape[i], value), prepend='The value(s) of')


def _check_type(name, dtype, types_):
    """
    Check the type of a variable.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    dtype : type
        The data type of the variable to be validated.
    types_ : set-like
        The accepted types.

    """

    valid = False

    for type_ in types_:
        if type_ not in _types.keys():
            _report(ValueError, 'must be in {!r}.', _types.keys(),
                    var_name='type_', var_value=type_,
                    prepend='Invalid validation call: ')

        if isinstance(dtype(), _types[type_][None]):
            valid = True
            break

    if not valid:
        _report(TypeError, '>>{}.dtype<<, {!r}, must be in {!r}.',
                (name, dtype, types_), prepend='The value(s) of ')


def _examine_var(name, var):
    """
    Examine a variable.

    The present function examines the data type, the bounds, and the shape of a
    variable. The variable can be of a built-in type, a numpy type, or a
    subclass of the `magni.utils.validation.types.MatrixBase` class.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    var : None
        The value of the variable to be examined.

    Returns
    -------
    dtype : type
        The data type of the examined variable.
    bounds : tuple
        The bounds of the examined variable.
    dshape : tuple
        The shape of the examined variable.

    """

    if isinstance(var, (np.generic, np.ndarray)):
        dtype = var.dtype.type

        if len(var.shape) > 0 and max(var.shape) == 0:
            bounds = ((np.inf, -np.inf), (np.inf, -np.inf))
        elif isinstance(dtype(), _types['complex'][None]):
            bounds = ((np.nanmin(var.real), np.nanmax(var.real)),
                      (np.nanmin(var.imag), np.nanmax(var.imag)))
        else:
            bounds = ((np.nanmin(var), np.nanmax(var)), (0, 0))

        dshape = var.shape
    elif isinstance(var, _MatrixBase):
        dtype = var.dtype
        bounds = var.bounds
        dshape = var.shape
    else:
        dtype = type(var)

        if isinstance(var, _types['complex'][None]):
            bounds = ((var.real, var.real), (var.imag, var.imag))
        else:
            bounds = ((var, var), (0, 0))

        dshape = ()

    if np.any(np.isnan(bounds[0])):
        bounds = ((np.inf, -np.inf), bounds[1])

    if np.any(np.isnan(bounds[1])):
        bounds = (bounds[0], (np.inf, -np.inf))

    return dtype, bounds, dshape



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions that may be used to chase data.

Routine listings
----------------
get_main_file_name()
    Function that returns the name of the main file/script.
get_main_file_source()
    Function that returns the source code of the main file/script.
get_main_source()
    Function that returns the 'local' source code of the main file/script.
get_stack_trace()
    Function that returns the complete stack trace.

"""

from __future__ import absolute_import  # stdlib io is shadowed by . io
from __future__ import division
import contextlib
import inspect
import traceback

try:
    from StringIO import StringIO  # Python 2 byte str (Python 2 only)
except ImportError:
    from io import StringIO  # Python 3 unicode str (both Py2 and Py3)

__all__ = ['get_main_file_name', 'get_main_file_source',
           'get_main_source', 'get_stack_trace']


def get_main_file_name():
    """
    Return the name of the main file/script which called this function.

    This will (if possible) return the name of the main file/script invoked by
    the Python interpreter i.e., the returned name is the name of the file, in
    which the bottom call on the stack is defined. Thus, the main file is still
    returned if the call to this function is buried deep in the code.

    Returns
    -------
    source_file : str
        The name of the main file/script which called this function.

    Notes
    -----
    It may not be possible to associate the call to this function with a main
    file. This is for instance the case if the call is handled by an IPython
    kernel. If no file is found, the returned string will indicate so.

    """

    stack_bottom = inspect.stack()[-1][0]

    try:
        source_file = inspect.getsourcefile(stack_bottom)
    except TypeError as e:
        source_file = 'Failed with TypeError: {!r}'.format(e.args[0])

    if source_file is None:
        source_file = 'Failed to find main file.'

    return source_file


def get_main_file_source():
    """
    Return the source code of the main file/script which called this function.

    This will (if possible) return the source code of the main file/script
    invoked by the Python interpreter i.e., the returned source code is the
    source code of the file, in which the bottom call on the stack is defined.
    Thus, the source code is still returned if the call to this function is
    buried deep in the code.

    Returns
    -------
    source : str
       The source code of the main file/script which called this function.

    Notes
    -----
    It may not be possible to associate the call to this function with a main
    file. This is for instance the case if the call is handled by an IPython
    kernel. If no file is found, the returned string will indicate so.

    """

    stack_bottom = inspect.stack()[-1][0]

    try:
        source_file = inspect.getsourcefile(stack_bottom)
    except TypeError as e:
        source_file = 'Failed with TypeError: {!r}'.format(e.args[0])

    if source_file is not None:
        try:
            with open(source_file) as sourcefile:
                source = sourcefile.read()
        except (OSError, IOError) as e:
            source = 'Failed with (OS/IO)Error: {!r}'.format(e.args[0])
    else:
        source = 'Failed to find main file.'

    return source


def get_main_source():
    """
    Return the local source code of the main file which called this function.

    This will (if possible) return the part of the source code of the main
    file/script surrounding (local to) the call to this function. The returned
    source code is a part of the source code of the file, in which the bottom
    call on the stack is defined. Thus, the source code is still returned if
    the call to this function is buried deep in the code.

    Returns
    -------
    source : str
        The local source code of the main file/script.

    Notes
    -----
    It may not be possible to associate the call to this function with a main
    file. This is for instance the case if the call is handled by an IPython
    kernel. If no file is found, the returned string will indicate so.

    """

    stack_bottom = inspect.stack()[-1][0]

    try:
        source = inspect.getsource(stack_bottom)
    except (OSError, IOError) as e:
        source = 'Failed with (OS/IO)Error: {!r}'.format(e.args[0])

    if source is None:
        source = 'Failed to find main file.'

    return source


def get_stack_trace():
    """
    Return the complete stack trace that led to the call to this function.

    A (pretty) print version of the stack trace is returned.

    Returns
    -------
    printed_stack : str
       The stack trace that led to the call to this function.

    """

    with contextlib.closing(StringIO()) as str_file:
        traceback.print_stack(file=str_file)  # Needs str (not unicode) on Py2
        printed_stack = str_file.getvalue()

    return printed_stack



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing functionality for image dictionary manipulations.

Routine listings
----------------
get_DCT(shape, overcomplete_shape=None)
    Get the DCT fast operation dictionary for the given image shape.
get_DFT(shape, overcomplete_shape=None)
    Get the DFT fast operation dictionary for the given image shape.
def get_DFT_transform_matrix(N)
    Return the normalised N-by-N discrete fourier transform (DFT) matrix.
analysis
    Module providing functionality to analyse dictionaries.
utils
    Module providing utility functions for the dictionaries subpackage.

"""

from magni.imaging.dictionaries._matrices import get_DCT, get_DFT
from magni.imaging.dictionaries._mtx1D import (
    get_DCT_transform_matrix, get_DFT_transform_matrix)
from magni.imaging.dictionaries import analysis
from magni.imaging.dictionaries import utils



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the `validate_generic` function.

Routine listings
----------------
validate_generic(name, type, value_in=None, len_=None, keys_in=None,
    has_keys=None, superclass=None, ignore_none=False, var=None)
    Validate non-numeric objects.

"""

from __future__ import absolute_import
from __future__ import division

import types

from magni.utils.validation._util import get_var as _get_var
from magni.utils.validation._util import report as _report


try:
    unicode = unicode
except NameError:
    _string = (str, bytes)
else:
    _string = (str, unicode)

_types = {'string': _string,
          'explicit collection': (list, tuple),
          # 'implicit collection': ...,
          # 'collection': ...,
          'mapping': dict,
          'function': types.FunctionType,
          'class': type}


def validate_generic(name, type_, value_in=None, len_=None, keys_in=None,
                     has_keys=None, superclass=None, ignore_none=False,
                     var=None):
    """
    Validate non-numeric objects.

    The present function is meant to validate the type or class of an object.
    Furthermore, if the object may only take on a limited number of values, the
    object can be validated against this list. In the case of collections (for
    example lists and tuples) and mappings (for example dictionaries), a
    specific length can be required. Furthermore, in the case of mappings, the
    keys can be validated by requiring and/or only allowing certain keys.

    If the present function is called with `name` set to None, an iterable with
    the value 'generic' followed by the remaining arguments passed in the call
    is returned. This is useful in combination with the validation function
    `magni.utils.validation.validate_levels`.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    type_ : None
        One or more references to groups of types, specific types, and/or
        specific classes.
    value_in : set-like
        The list of values accepted. (the default is None, which implies that
        all values are accepted)
    len_ : int
        The length required. (the default is None, which implies that all
        lengths are accepted)
    keys_in : set-like
        The list of accepted keys. (the default is None, which implies that all
        keys are accepted)
    has_keys : set-like
        The list of required keys. (the default is None, which implies that no
        keys are required)
    superclass : class
        The required superclass. (the default is None, which implies that no
        superclass is required)
    ignore_none : bool
        A flag indicating if the variable is allowed to be none. (the default
        is False)
    var : None
        The value of the variable to be validated.

    See Also
    --------
    magni.utils.validation.validate_levels : Validate contained objects.
    magni.utils.validation.validate_numeric : Validate numeric objects.

    Notes
    -----
    `name` must refer to a variable in the parent scope of the function or
    method decorated by `magni.utils.validation.decorate_validation` which is
    closest to the top of the call stack. If `name` is a string then there must
    be a variable of that name in that scope. If `name` is a set-like object
    then there must be a variable having the first value in that set-like
    object as name. The remaining values are used as keys/indices on the
    variable to obtain the variable to be validated. For example, the `name`
    ('name', 0, 'key') refers to the variable "name[0]['key']".

    `type_` is either a single value treated as a list with one value or a
    set-like object containing at least one value. Each value is either a
    specific type or class, or it refers to one or more types by having one of
    the string values 'string', 'explicit collection', 'implicict collection',
    'collection', 'mapping', 'function', 'class'.

    - 'string' tests if the variable is a str.
    - 'explicit collection' tests if the variable is a list or tuple.
    - 'implicit collection' tests if the variable is iterable.
    - 'collection' is a combination of the two above.
    - 'mapping' tests if the variable is a dict.
    - 'function' tests if the variable is a function.
    - 'class' tests if the variable is a type.

    `var` can be used to pass the value of the variable to be validated. This
    is useful either when the variable cannot be looked up by `name` (for
    example, if the variable is a property of the argument of a function) or to
    remove the overhead of looking up the value.

    Examples
    --------
    Every public function and method of the present package (with the exception
    of the functions of this subpackage itself) validates every argument and
    keyword argument using the functionality of this subpackage. Thus, for
    examples of how to use the present function, browse through the code.

    """

    if name is None:
        return ('generic', type_, value_in, len_, keys_in, has_keys,
                superclass, ignore_none)

    if var is None:
        var = _get_var(name)

    if var is None:
        if not isinstance(ignore_none, bool):
            _report(TypeError, 'must be {!r}.', bool, var_name='ignore_none',
                    var_value=ignore_none, expr='type({})',
                    prepend='Invalid validation call: ')

        if ignore_none:
            return
        else:
            _report(ValueError, 'must not be {!r}.', None, var_name=name)

    if isinstance(type_, (str, type)) or not hasattr(type_, '__iter__'):
        type_ = (type_,)

    _check_type(name, var, type_)
    _check_value(name, var, value_in)
    _check_len(name, var, len_)
    _check_keys(name, var, keys_in, has_keys)
    _check_inheritance(name, var, superclass)


def _check_inheritance(name, var, superclass):
    """
    Check the superclasses of a variable.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    var : None
        The value of the variable to be validated.
    superclass : class
        The required superclass.

    """

    if superclass is not None:
        if not isinstance(superclass, type):
            _report(TypeError, 'must be a type.', var_name='superclass',
                    var_value=superclass, prepend='Invalid validation call: ')

        if not issubclass(var, superclass):
            _report(TypeError, 'must be a subclass of {!r}.', superclass,
                    var_name=name)


def _check_keys(name, var, keys_in, has_keys):
    """
    Check the keys of a variable.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    var : None
        The value of the variable to be validated.
    keys_in : set-like
        The allowed keys.
    has_keys : set-like
        The required keys.

    """

    if keys_in is not None:
        if not isinstance(keys_in, (list, tuple)):
            _report(TypeError, 'must be in {!r}.', (list, tuple),
                    var_name='keys_in', var_value=keys_in,
                    prepend='Invalid validation call: ')

        for key in var.keys():
            if key not in keys_in:
                _report(KeyError, 'must be in {!r}.', keys_in, var_name=name,
                        expr='{}.keys()')

    if has_keys is not None:
        if not isinstance(has_keys, (list, tuple)):
            _report(TypeError, 'must be in {!r}.', (list, tuple),
                    var_name='has_keys', var_value=has_keys,
                    prepend='Invalid validation call: ')

        for key in has_keys:
            if key not in var.keys():
                _report(KeyError, 'must have the key {!r}.', key,
                        var_name=name)


def _check_len(name, var, len_):
    """
    Check the length of a variable.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    var : None
        The value of the variable to be validated.
    len_ : int
        The required length.

    """

    if len_ is not None:
        if not isinstance(len_, int):
            _report(TypeError, 'must be {!r}.', int, var_name='len_',
                    var_value=len_, expr='type({})',
                    prepend='Invalid validation call: ')

        if not hasattr(var, '__len__') or len(var) != len_:
            _report(ValueError, 'must be {!r}.', len_, var_name=name,
                    expr='len({})')


def _check_type(name, var, types_):
    """
    Check the type of a variable.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    var : None
        The value of the variable to be validated.
    types_ : set-like
        The allowed types.

    """

    valid = False

    for type_ in types_:
        if isinstance(type_, type):
            if isinstance(var, type_):
                valid = True
                break
        elif type_ in _types.keys():
            if isinstance(var, _types[type_]):
                valid = True
                break
        elif type_ in ('implicit collection', 'collection'):
            if hasattr(var, '__iter__'):
                valid = True
                break
        else:
            _report(ValueError, 'must be a valid type.', var_name='type_',
                    var_value=type_, prepend='Invalid validation call: ')

    if not valid:
        _report(TypeError, 'must be in {!r}.', types_, var_name=name,
                expr='type({})')


def _check_value(name, var, value_in):
    """
    Check the value of a variable.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    var : None
        The value of the variable to be validated.
    value_in : set-like
        The allowed values.

    """

    if value_in is not None:
        if not isinstance(value_in, (list, tuple)):
            _report(TypeError, 'must be in {!r}.', (list, tuple),
                    var_name='value_in', var_value=value_in,
                    prepend='Invalid validation call: ')

        if var not in value_in:
            _report(ValueError, 'must be in {!r}.', value_in, var_name=name)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the common functionality of the subpackage.

"""

from __future__ import division
from collections import OrderedDict as _OrderedDict

from magni.utils.types import ClassProperty as _classproperty
from magni.utils.types import ReadOnlyDict as _ReadOnlyDict
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels


class BaseClass(object):
    """
    Base class of every `magni.afm.types` data class.

    The present class validates the attributes passed to its constructor
    against the allowed attributes of the class of the given instance and
    exposes these attributes through a read-only dictionary property, `attrs`.
    Furthermore, the present class exposes the allowed attributes of the class
    of the given instance through a read-only dictionary static property,
    `params`.

    Parameters
    ----------
    attrs : dict
        The desired attributes of the instance.

    Attributes
    ----------
    attrs : magni.utils.types.ReadOnlyDict
        The attributes of the instance.
    params : magni.utils.types.ReadOnlyDict
        The allowed attributes of the instance.

    Examples
    --------
    An example could be the subclass, 'Person' which allows only the string
    attribute, name:

    >>> from magni.afm.types._util import BaseClass
    >>> class Person(BaseClass):
    ...     def __init__(self, attrs):
    ...         BaseClass.__init__(self, attrs)
    ...     _params = {'name': str}

    This class can then be initiated with a name:

    >>> person = Person({'name': 'Murphy'})
    >>> print('{!r}'.format(person.attrs['name']))
    'Murphy'

    Any other attributes, than 'name', passed to the class are ignored:

    >>> person = Person({'name': 'Murphy', 'age': 42})
    >>> for name in person.attrs.keys():
    ...     print('{!r}'.format(name))
    'name'

    """

    def __init__(self, attrs):
        @_decorate_validation
        def validate_input():
            _generic('attrs', 'mapping')
            params = self.params

            for name in attrs.keys():
                if name in params.keys():
                    _generic(('attrs', name), params[name])

        validate_input()

        names = self.params.keys()
        attrs = {name: value for name, value in attrs.items() if name in names}
        self._attrs = _ReadOnlyDict(attrs)

    attrs = property(lambda self: self._attrs)

    @_classproperty
    def params(class_):
        """
        Get the allowed attributes of the class (of the instance).

        Which attributes are allowed is specified by the class and superclasses
        (of the instance).

        Returns
        -------
        params : magni.utils.types.ReadOnlyDict
            The allowed attributes.

        Notes
        -----
        The allowed attributes of the class (of the given instance) are found
        by inspecting the class and its base classes for the static variable
        '_params' and collecting these in a read-only dictionary.

        """

        params = []

        while class_ != BaseClass:
            if hasattr(class_, '_params'):
                params.extend(class_._params.items())

            class_ = class_.__base__

        return _ReadOnlyDict(params)


class File(BaseClass):
    """
    Base class of the `magni.afm.types` file classes.

    The present class specifies the allowed file-level attributes and exposes
    the read-only property, buffers which all .mi files have. Furthermore, the
    present class provides a method for accessing buffers by their
    'bufferLabel' attribute.

    Parameters
    ----------
    attrs : list or tuple
        The desired attributes of the file.
    buffers : list or tuple
        The buffers of the file.

    Attributes
    ----------
    buffers : tuple
        The buffers of the file.

    See Also
    --------
    BaseClass : Superclass of the present class.

    Examples
    --------
    A subclass of the present class is implicitly instantiated when loading,
    for example, the .mi file provided with the package:

    >>> import os, magni
    >>> path = magni.utils.split_path(magni.__path__[0])[0]
    >>> path = path + 'examples' + os.sep + 'example.mi'
    >>> if os.path.isfile(path):
    ...     file_ = magni.afm.io.read_mi_file(path)

    This file has a number of buffers which each has the 'bufferLabel'
    attribute:

    >>> if os.path.isfile(path):
    ...     for buffer_ in file_.buffers[::2][:3]:
    ...         label = buffer_.attrs['bufferLabel']
    ...         print("Buffer with 'bufferLabel': {!r}".format(label))
    ... else:
    ...     for label in ('Topography', 'Deflection', 'Friction'):
    ...         print("Buffer with 'bufferLabel': {!r}".format(label))
    Buffer with 'bufferLabel': 'Topography'
    Buffer with 'bufferLabel': 'Deflection'
    Buffer with 'bufferLabel': 'Friction'

    If only, for example, buffers with 'bufferLabel' equal to 'Topography' are
    desired, the method, `get_buffer` can be called:

    >>> if os.path.isfile(path):
    ...     buffers = len(file_.get_buffer('Topography'))
    ...     print("Buffers with 'bufferLabel' == 'Topography': {}"
    ...            .format(buffers))
    ... else:
    ...     print("Buffers with 'bufferLabel' == 'Topography': 2")
    Buffers with 'bufferLabel' == 'Topography': 2

    """

    _params = _OrderedDict((('fileType', str),
                            ('dateAcquired', str),
                            ('xOffset', float),
                            ('yOffset', float),
                            ('xLength', float),
                            ('yLength', float),
                            ('data', str)))

    def __init__(self, attrs, buffers):
        @_decorate_validation
        def validate_input():
            _generic('attrs', 'mapping', has_keys=('fileType',))
            _levels('buffers', (_generic(None, 'explicit collection'),
                                _generic(None, BaseClass)))

        BaseClass.__init__(self, attrs)
        validate_input()

        self._buffers = tuple(buffers)

    buffers = property(lambda self: self._buffers)

    def get_buffer(self, label):
        """
        Get the buffers that have the specified buffer label.

        Parameters
        ----------
        label : str
            The desired buffer label of the buffer.

        Returns
        -------
        buffers : list
            The buffers that have the desired buffer label.

        """

        @_decorate_validation
        def validate_input():
            _generic('label', 'string')

        validate_input()

        buffers = []

        for buffer_ in self.buffers:
            if ('bufferLabel' in buffer_.attrs.keys() and
                    buffer_.attrs['bufferLabel'] == label):
                buffers.append(buffer_)

        return buffers


class FileCollection(BaseClass):
    """
    Data class for collections of File instances with identical settings.

    The settings are the following attributes: 'fileType', 'mode', 'xPixels',
    'yPixels', 'xOffset', 'yOffset', 'xLength', 'yLength', 'scanSpeed',
    'acMac', 'acACMode', 'plotType'.

    The present class exposes the files of the collection through a read-only
    tuple property, `files` and the paths of these files through a read-only
    tuple property, `paths`.

    Parameters
    ----------
    files : list or tuple
        The files of the collection.
    paths : list or tuple
        The paths of the files of the collection.

    Attributes
    ----------
    files : tuple
        The files of the collection.
    paths : tuple
        The paths of the files of the collection.

    See Also
    --------
    BaseClass : Superclass of the present class.

    Examples
    --------
    No example .mi file collection is distributed with `magni`.

    """

    _params = _OrderedDict((('fileType', str),
                            ('mode', str),
                            ('xPixels', int),
                            ('yPixels', int),
                            ('xOffset', float),
                            ('yOffset', float),
                            ('xLength', float),
                            ('yLength', float),
                            ('scanSpeed', float),
                            ('acMac', bool),
                            ('acACMode', bool),
                            ('plotType', str)))

    def __init__(self, files, paths):
        @_decorate_validation
        def validate_input():
            _levels('files', (_generic(None, 'explicit collection'),
                              _generic(None, File)))

            if len(files) < 1:
                msg = 'The value of >>len(files)<<, {!r}, must be > 0.'
                raise ValueError(msg.format(len(files)))

            attrs = [{name: value for name, value in file_.attrs.items()
                      if name in self._params} for file_ in files]

            for i in range(1, len(files)):
                if attrs[i] != attrs[0]:
                    raise ValueError('The values of >>files<< must have the '
                                     'same settings.')

            _levels('paths', (_generic(None, 'explicit collection',
                                       len_=len(files)),
                              _generic(None, 'string')))

        # validate before calling the superclass constructor to ensure that
        # 'files[0].attrs' exists
        validate_input()
        BaseClass.__init__(self, files[0].attrs)

        self._files = tuple(files)
        self._paths = tuple(paths)

    files = property(lambda self: self._files)
    paths = property(lambda self: self._paths)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing intuitive and extensive multiprocessing functionality.

Routine listings
----------------
config
    Configger providing configuration options for this subpackage.
File()
    Control pytables access to hdf5 files when using multiprocessing.
process(func, namespace={}, args_list=None, kwargs_list=None, maxtasks=None)
    Map multiple function calls to multiple processors.

Notes
-----
See `_util` for documentation of `File`.
See `_processing` for documentation of `process`.

"""

from magni.utils.multiprocessing._config import configger as config
from magni.utils.multiprocessing._util import File
from magni.utils.multiprocessing._processing import process
